from __future__ import annotations

import tempfile
from pathlib import Path

import pytest

from uvscem import vscode_paths


@pytest.fixture
def _home(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> Path:
    monkeypatch.setattr(Path, "home", lambda: tmp_path)
    return tmp_path


def test_detect_runtime_environment_uses_explicit_override(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setenv("UVSCEM_RUNTIME", "custom-env")

    assert vscode_paths.detect_runtime_environment() == "custom-env"


def test_detect_runtime_environment_uses_vscode_agent_folder(
    monkeypatch: pytest.MonkeyPatch,
    _home: Path,
) -> None:
    monkeypatch.delenv("UVSCEM_RUNTIME", raising=False)
    monkeypatch.setenv(
        "VSCODE_AGENT_FOLDER", str(Path(tempfile.gettempdir()) / "vscode-agent")
    )

    assert vscode_paths.detect_runtime_environment() == "vscode-remote"


def test_detect_runtime_environment_prefers_vscode_server_home(
    monkeypatch: pytest.MonkeyPatch,
    _home: Path,
) -> None:
    monkeypatch.delenv("UVSCEM_RUNTIME", raising=False)
    monkeypatch.delenv("VSCODE_AGENT_FOLDER", raising=False)
    _home.joinpath(".vscode-server").mkdir(parents=True)

    assert vscode_paths.detect_runtime_environment() == "vscode-server"


def test_detect_runtime_environment_prefers_vscode_remote_home(
    monkeypatch: pytest.MonkeyPatch,
    _home: Path,
) -> None:
    monkeypatch.delenv("UVSCEM_RUNTIME", raising=False)
    monkeypatch.delenv("VSCODE_AGENT_FOLDER", raising=False)
    _home.joinpath(".vscode-remote").mkdir(parents=True)

    assert vscode_paths.detect_runtime_environment() == "vscode-remote"


def test_detect_runtime_environment_defaults_to_local(
    monkeypatch: pytest.MonkeyPatch,
    _home: Path,
) -> None:
    monkeypatch.delenv("UVSCEM_RUNTIME", raising=False)
    monkeypatch.delenv("VSCODE_AGENT_FOLDER", raising=False)

    assert vscode_paths.detect_runtime_environment() == "local"


def test_resolve_vscode_root_uses_explicit_root(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    custom_root = Path(tempfile.gettempdir()) / "custom-vscode"
    monkeypatch.setenv("UVSCEM_VSCODE_ROOT", str(custom_root))

    assert vscode_paths.resolve_vscode_root() == custom_root.resolve()


def test_resolve_vscode_root_uses_vscode_agent_folder(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.delenv("UVSCEM_VSCODE_ROOT", raising=False)
    agent_root = Path(tempfile.gettempdir()) / "vscode-agent"
    monkeypatch.setenv("VSCODE_AGENT_FOLDER", str(agent_root))

    assert vscode_paths.resolve_vscode_root() == agent_root.resolve()


def test_resolve_vscode_root_uses_runtime_environment(
    monkeypatch: pytest.MonkeyPatch,
    _home: Path,
) -> None:
    monkeypatch.delenv("UVSCEM_VSCODE_ROOT", raising=False)
    monkeypatch.delenv("VSCODE_AGENT_FOLDER", raising=False)
    monkeypatch.setattr(
        vscode_paths, "detect_runtime_environment", lambda: "vscode-server"
    )

    assert vscode_paths.resolve_vscode_root() == _home.joinpath(".vscode-server")


def test_resolve_vscode_root_uses_runtime_environment_remote(
    monkeypatch: pytest.MonkeyPatch,
    _home: Path,
) -> None:
    monkeypatch.delenv("UVSCEM_VSCODE_ROOT", raising=False)
    monkeypatch.delenv("VSCODE_AGENT_FOLDER", raising=False)
    monkeypatch.setattr(
        vscode_paths, "detect_runtime_environment", lambda: "vscode-remote"
    )

    assert vscode_paths.resolve_vscode_root() == _home.joinpath(".vscode-remote")


def test_resolve_vscode_root_uses_existing_candidate(
    monkeypatch: pytest.MonkeyPatch,
    _home: Path,
) -> None:
    monkeypatch.delenv("UVSCEM_VSCODE_ROOT", raising=False)
    monkeypatch.delenv("VSCODE_AGENT_FOLDER", raising=False)
    monkeypatch.setattr(vscode_paths, "detect_runtime_environment", lambda: "local")
    _home.joinpath(".vscode").mkdir(parents=True)

    assert vscode_paths.resolve_vscode_root() == _home.joinpath(".vscode")


def test_resolve_vscode_root_linux_fallback(
    monkeypatch: pytest.MonkeyPatch,
    _home: Path,
) -> None:
    monkeypatch.delenv("UVSCEM_VSCODE_ROOT", raising=False)
    monkeypatch.delenv("VSCODE_AGENT_FOLDER", raising=False)
    monkeypatch.setattr(vscode_paths, "detect_runtime_environment", lambda: "local")
    monkeypatch.setattr(vscode_paths.platform, "system", lambda: "Linux")

    assert vscode_paths.resolve_vscode_root() == _home.joinpath(".vscode-server")


def test_resolve_vscode_root_non_linux_fallback(
    monkeypatch: pytest.MonkeyPatch,
    _home: Path,
) -> None:
    monkeypatch.delenv("UVSCEM_VSCODE_ROOT", raising=False)
    monkeypatch.delenv("VSCODE_AGENT_FOLDER", raising=False)
    monkeypatch.setattr(vscode_paths, "detect_runtime_environment", lambda: "local")
    monkeypatch.setattr(vscode_paths.platform, "system", lambda: "Darwin")

    assert vscode_paths.resolve_vscode_root() == _home.joinpath(".vscode")



from __future__ import annotations

import argparse
import base64
import hashlib
import io
import os
import platform
import shutil
import stat
import tarfile
import tempfile
import uuid
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Iterator
from urllib.parse import quote

import requests

DEFAULT_VSCE_SIGN_VERSION = "2.0.6"
PACKAGE_PREFIX = "@vscode/vsce-sign-"
SUPPORTED_VSCE_SIGN_TARGETS = (
    "linux-x64",
    "linux-arm64",
    "linux-arm",
    "alpine-x64",
    "alpine-arm64",
    "darwin-x64",
    "darwin-arm64",
    "win32-x64",
    "win32-arm64",
)


class VsceSignBootstrapError(RuntimeError):
    """Raised when vsce-sign binary bootstrap fails."""


@dataclass
class VsceSignRunInstallation:
    """Represents one vsce-sign installation used for a full program run."""

    binary_path: Path
    temporary_dir: Path | None = None

    def cleanup(self) -> None:
        """Remove temporary install directory, if this run used one."""
        if self.temporary_dir and self.temporary_dir.exists():
            shutil.rmtree(self.temporary_dir, ignore_errors=True)


def _is_musl_linux() -> bool:
    libc_name, _ = platform.libc_ver()
    if libc_name.lower() == "musl":
        return True
    return Path("/etc/alpine-release").is_file()


def get_vsce_sign_target() -> str:
    """Return the vsce-sign platform target string used by npm packages."""
    system = platform.system().lower()
    machine = platform.machine().lower()

    if system == "linux":
        if machine in {"x86_64", "amd64"}:
            return "alpine-x64" if _is_musl_linux() else "linux-x64"
        if machine in {"aarch64", "arm64"}:
            return "alpine-arm64" if _is_musl_linux() else "linux-arm64"
        if machine in {"armv7l", "armv6l", "arm"}:
            return "linux-arm"
        raise VsceSignBootstrapError(f"Unsupported Linux architecture: {machine}")

    if system == "darwin":
        if machine in {"x86_64", "amd64"}:
            return "darwin-x64"
        if machine in {"arm64", "aarch64"}:
            return "darwin-arm64"
        raise VsceSignBootstrapError(f"Unsupported macOS architecture: {machine}")

    if system == "windows":
        if machine in {"x86_64", "amd64"}:
            return "win32-x64"
        if machine in {"arm64", "aarch64"}:
            return "win32-arm64"
        raise VsceSignBootstrapError(f"Unsupported Windows architecture: {machine}")

    raise VsceSignBootstrapError(f"Unsupported operating system: {system}")


def get_vsce_sign_package_name(target: str | None = None) -> str:
    """Return the platform-specific npm package name for vsce-sign."""
    resolved_target = target if target is not None else get_vsce_sign_target()
    return f"{PACKAGE_PREFIX}{resolved_target}"


def _verify_npm_integrity(tarball_bytes: bytes, integrity: str) -> None:
    if not integrity.startswith("sha512-"):
        raise VsceSignBootstrapError(
            f"Unsupported npm integrity algorithm: {integrity}"
        )

    expected = integrity.split("-", maxsplit=1)[1]
    actual = base64.b64encode(hashlib.sha512(tarball_bytes).digest()).decode("ascii")
    if expected != actual:
        raise VsceSignBootstrapError(
            "Integrity verification failed while downloading vsce-sign binary package"
        )


def _fetch_package_info(
    package_name: str,
    version: str,
    session: Any,
) -> tuple[str, str]:
    encoded_name = quote(package_name, safe="")
    response = session.get(f"https://registry.npmjs.org/{encoded_name}", timeout=30)
    response.raise_for_status()

    metadata = response.json()
    version_data = metadata.get("versions", {}).get(version)
    if not version_data:
        raise VsceSignBootstrapError(
            f"Version {version} not found for package {package_name}"
        )

    dist = version_data.get("dist", {})
    tarball_url = str(dist.get("tarball", ""))
    integrity = str(dist.get("integrity", ""))
    if not tarball_url or not integrity:
        raise VsceSignBootstrapError(
            f"Missing tarball metadata for package {package_name}@{version}"
        )
    return tarball_url, integrity


def _binary_name_for_target(target: str) -> str:
    return "vsce-sign.exe" if target.startswith("win32-") else "vsce-sign"


def _binary_name() -> str:
    return _binary_name_for_target(get_vsce_sign_target())


def _extract_binary_bytes_from_tarball(
    tarball_bytes: bytes,
    binary_name: str,
) -> bytes:
    archive_member = f"package/bin/{binary_name}"
    with tarfile.open(fileobj=io.BytesIO(tarball_bytes), mode="r:gz") as archive:
        try:
            member = archive.getmember(archive_member)
        except KeyError as exc:
            raise VsceSignBootstrapError(
                f"Expected binary {archive_member} not found in package archive"
            ) from exc
        extracted = archive.extractfile(member)
        if extracted is None:
            raise VsceSignBootstrapError(
                f"Expected binary {archive_member} not found in package archive"
            )
        return extracted.read()


def _expected_binary_sha512(
    package_name: str,
    version: str,
    binary_name: str,
    session: Any,
) -> str:
    tarball_url, integrity = _fetch_package_info(package_name, version, session)
    tarball_response = session.get(tarball_url, timeout=60)
    tarball_response.raise_for_status()
    tarball_bytes = tarball_response.content
    _verify_npm_integrity(tarball_bytes, integrity)
    expected_bytes = _extract_binary_bytes_from_tarball(tarball_bytes, binary_name)
    return hashlib.sha512(expected_bytes).hexdigest()


def install_vsce_sign_binary(
    install_dir: str | Path,
    version: str = DEFAULT_VSCE_SIGN_VERSION,
    force: bool = False,
    session: Any | None = None,
    verify_existing_checksum: bool = True,
) -> Path:
    """Install platform-specific vsce-sign binary without requiring Node.js."""
    target = get_vsce_sign_target()
    return install_vsce_sign_binary_for_target(
        target=target,
        install_dir=install_dir,
        version=version,
        force=force,
        session=session,
        verify_existing_checksum=verify_existing_checksum,
    )


def install_vsce_sign_binary_for_target(
    target: str,
    install_dir: str | Path,
    version: str = DEFAULT_VSCE_SIGN_VERSION,
    force: bool = False,
    session: Any | None = None,
    verify_existing_checksum: bool = True,
) -> Path:
    """Install vsce-sign binary for a specific npm platform target."""
    package_name = get_vsce_sign_package_name(target)
    install_path = Path(install_dir).expanduser().resolve()
    binary_name = _binary_name_for_target(target)
    binary_path = install_path.joinpath(binary_name)
    session_client: Any = session if session is not None else requests.Session()

    if binary_path.is_file() and not force:
        if not verify_existing_checksum:
            return binary_path
        expected_digest = _expected_binary_sha512(
            package_name=package_name,
            version=version,
            binary_name=binary_name,
            session=session_client,
        )
        actual_digest = hashlib.sha512(binary_path.read_bytes()).hexdigest()
        if actual_digest == expected_digest:
            return binary_path
        force = True

    install_path.mkdir(parents=True, exist_ok=True)
    tarball_url, integrity = _fetch_package_info(package_name, version, session_client)
    tarball_response = session_client.get(tarball_url, timeout=60)
    tarball_response.raise_for_status()
    tarball_bytes = tarball_response.content
    _verify_npm_integrity(tarball_bytes, integrity)
    binary_bytes = _extract_binary_bytes_from_tarball(tarball_bytes, binary_name)

    with NamedTemporaryFile(delete=False, dir=install_path) as tmp_file:
        tmp_target = Path(tmp_file.name)

    try:
        with open(tmp_target, "wb") as output:
            output.write(binary_bytes)

        if binary_name == "vsce-sign":
            current_mode = os.stat(tmp_target).st_mode
            os.chmod(
                tmp_target,
                current_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH,
            )
        tmp_target.replace(binary_path)
    finally:
        if tmp_target.exists() and tmp_target != binary_path:
            tmp_target.unlink(missing_ok=True)

    return binary_path


def install_vsce_sign_binary_with_fallback(
    install_dir: str | Path,
    version: str = DEFAULT_VSCE_SIGN_VERSION,
    force: bool = False,
    session: Any | None = None,
    temp_root: str | Path | None = None,
    verify_existing_checksum: bool = True,
) -> VsceSignRunInstallation:
    """Install vsce-sign with fallback to UUID temp dir when primary path is not writable."""
    try:
        binary_path = install_vsce_sign_binary(
            install_dir=install_dir,
            version=version,
            force=force,
            session=session,
            verify_existing_checksum=verify_existing_checksum,
        )
        return VsceSignRunInstallation(binary_path=binary_path)
    except PermissionError:
        pass

    base_temp_dir = (
        Path(temp_root).expanduser().resolve()
        if temp_root is not None
        else Path(tempfile.gettempdir()).resolve()
    )
    fallback_dir = base_temp_dir.joinpath(f"uvscem-vsce-sign-{uuid.uuid4()}")
    binary_path = install_vsce_sign_binary(
        install_dir=fallback_dir,
        version=version,
        force=True,
        session=session,
        verify_existing_checksum=False,
    )
    return VsceSignRunInstallation(binary_path=binary_path, temporary_dir=fallback_dir)


@contextmanager
def provision_vsce_sign_binary_for_run(
    install_dir: str | Path,
    version: str = DEFAULT_VSCE_SIGN_VERSION,
    force: bool = False,
    session: Any | None = None,
    temp_root: str | Path | None = None,
    verify_existing_checksum: bool = True,
) -> Iterator[Path]:
    """Provision one vsce-sign binary for a full run and always clean up temp fallback."""
    installation = install_vsce_sign_binary_with_fallback(
        install_dir=install_dir,
        version=version,
        force=force,
        session=session,
        temp_root=temp_root,
        verify_existing_checksum=verify_existing_checksum,
    )
    try:
        yield installation.binary_path
    finally:
        installation.cleanup()


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="uvscem-vsce-sign-bootstrap")
    parser.add_argument("--install-dir", default="~/.local/bin")
    parser.add_argument("--version", default=DEFAULT_VSCE_SIGN_VERSION)
    parser.add_argument("--force", action="store_true")
    return parser


def main() -> None:
    args = build_parser().parse_args()
    target = install_vsce_sign_binary(
        install_dir=args.install_dir,
        version=args.version,
        force=args.force,
    )
    print(target)


if __name__ == "__main__":
    main()



from __future__ import annotations

import os
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import Any

import requests


def stream_download_to_target(
    *,
    session: Any,
    url: str,
    target_path: Path,
    headers: dict[str, str],
    temp_prefix: str,
    timeout: tuple[int, int],
) -> Path:
    with tempfile.TemporaryDirectory(prefix=temp_prefix) as tmp_dir:
        file_path = Path(tmp_dir, target_path.name)

        with open(file_path, "wb") as output:
            response: requests.Response = session.get(
                url,
                stream=True,
                headers=headers,
                timeout=timeout,
            )
            response.raise_for_status()

            for chunk in response.iter_content(chunk_size=1024 * 8):
                if chunk:
                    output.write(chunk)
                    output.flush()
                    os.fsync(output.fileno())

        shutil.move(file_path, target_path)
        return target_path


def run_vsce_sign_verify(
    *,
    vsce_sign_binary: Path,
    extension_path: Path,
    signature_archive_path: Path,
    run_command: Any = subprocess.run,
) -> None:
    cmd = [
        str(vsce_sign_binary),
        "verify",
        "--package",
        str(extension_path),
        "--signaturearchive",
        str(signature_archive_path),
    ]
    process = run_command(
        cmd,
        capture_output=True,
        check=False,
        text=True,
    )
    if process.returncode != 0:
        raise subprocess.CalledProcessError(
            process.returncode,
            cmd,
            process.stdout,
            process.stderr,
        )


def run_code_cli_install(
    *,
    code_binary: str,
    extension_path: Path,
    run_command: Any = subprocess.run,
) -> None:
    cmd = [
        code_binary,
        "--install-extension",
        f"{extension_path}",
        "--force",
    ]
    output = run_command(
        cmd,
        capture_output=True,
        check=True,
        text=True,
    )
    error_msg = "Error: "
    if error_msg in f"{output.stdout}" or error_msg in f"{output.stderr}":
        raise subprocess.CalledProcessError(1, cmd, output.stdout, output.stderr)



from __future__ import annotations

import asyncio
import builtins
import os
import socket
import tempfile
from pathlib import Path

import pytest

from uvscem import code_manager
from uvscem.code_manager import CodeManager


def test_init_calls_socket_and_code_discovery(monkeypatch: pytest.MonkeyPatch) -> None:
    manager = CodeManager()

    assert manager.socket_path is None
    assert manager.code_path is None


def test_find_socket_removes_stale_and_sets_environment(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    stale = tmp_path / "vscode-ipc-stale.sock"
    active = tmp_path / "vscode-ipc-active.sock"
    stale.write_text("stale")
    active.write_text("active")

    manager = CodeManager.__new__(CodeManager)
    manager.socket_path = None
    manager.code_path = None

    monkeypatch.delenv("VSCODE_IPC_HOOK_CLI", raising=False)
    monkeypatch.setenv("XDG_RUNTIME_DIR", str(tmp_path))
    monkeypatch.setattr(code_manager.platform, "system", lambda: "Linux")
    monkeypatch.setattr(
        CodeManager,
        "is_socket_closed",
        staticmethod(lambda path: path.name.endswith("stale.sock")),
    )

    asyncio.run(manager.find_socket(update_environment=True))

    assert not stale.exists()
    assert manager.socket_path == active
    assert os.environ["VSCODE_IPC_HOOK_CLI"] == str(active)


def test_find_socket_without_any_socket_keeps_environment_unchanged(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    manager = CodeManager.__new__(CodeManager)
    manager.socket_path = None
    manager.code_path = None

    monkeypatch.delenv("VSCODE_IPC_HOOK_CLI", raising=False)
    monkeypatch.setenv("XDG_RUNTIME_DIR", str(tmp_path))
    monkeypatch.setattr(
        CodeManager,
        "is_socket_closed",
        staticmethod(lambda _path: False),
    )
    monkeypatch.delenv("VSCODE_IPC_HOOK_CLI", raising=False)

    asyncio.run(manager.find_socket(update_environment=True))

    assert manager.socket_path is None
    assert "VSCODE_IPC_HOOK_CLI" not in os.environ


def test_find_socket_uses_existing_hook_when_present(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    manager = CodeManager.__new__(CodeManager)
    manager.socket_path = None
    manager.code_path = None

    existing_hook = str(Path(tempfile.gettempdir()) / "vscode-existing.sock")
    monkeypatch.setenv("VSCODE_IPC_HOOK_CLI", existing_hook)
    asyncio.run(manager.find_socket(update_environment=True))

    assert manager.socket_path == Path(existing_hook)
    assert os.environ["VSCODE_IPC_HOOK_CLI"] == existing_hook


def test_find_socket_uses_existing_hook_without_environment_update(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    manager = CodeManager.__new__(CodeManager)
    manager.socket_path = None
    manager.code_path = None

    existing_hook = str(Path(tempfile.gettempdir()) / "vscode-existing.sock")
    monkeypatch.setenv("VSCODE_IPC_HOOK_CLI", existing_hook)
    asyncio.run(manager.find_socket(update_environment=False))

    assert manager.socket_path == Path(existing_hook)


def test_find_socket_windows_without_hook_returns_early(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    manager = CodeManager.__new__(CodeManager)
    manager.socket_path = None
    manager.code_path = None

    monkeypatch.delenv("VSCODE_IPC_HOOK_CLI", raising=False)
    monkeypatch.setenv("XDG_RUNTIME_DIR", str(tmp_path))
    monkeypatch.setattr(code_manager.platform, "system", lambda: "Windows")

    asyncio.run(manager.find_socket(update_environment=True))

    assert manager.socket_path is None


def test_find_socket_keeps_first_active_when_multiple_active(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    first = tmp_path / "vscode-ipc-aaa.sock"
    second = tmp_path / "vscode-ipc-bbb.sock"
    first.write_text("a")
    second.write_text("b")

    manager = CodeManager.__new__(CodeManager)
    manager.socket_path = None
    manager.code_path = None

    monkeypatch.delenv("VSCODE_IPC_HOOK_CLI", raising=False)
    monkeypatch.setenv("XDG_RUNTIME_DIR", str(tmp_path))
    monkeypatch.setattr(code_manager.platform, "system", lambda: "Linux")
    monkeypatch.setattr(
        CodeManager, "is_socket_closed", staticmethod(lambda _path: False)
    )

    asyncio.run(manager.find_socket(update_environment=False))

    assert manager.socket_path == first


def _write_code_launcher(script_path: Path, version: str, commit: str) -> None:
    script_path.parent.mkdir(parents=True, exist_ok=True)
    script_path.write_text(f"VERSION={version}\nCOMMIT={commit}\n")


def test_find_latest_code_updates_path(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path / "vscode"
    _write_code_launcher(
        root / "bin" / "commit1" / "bin" / "remote-cli" / "code", "1.0.0", "commit1"
    )
    _write_code_launcher(
        root / "bin" / "commit2" / "bin" / "remote-cli" / "code", "2.0.0", "commit2"
    )

    manager = CodeManager.__new__(CodeManager)
    manager.socket_path = None
    manager.code_path = None

    monkeypatch.setattr(code_manager, "vscode_root", root)
    monkeypatch.setattr(
        code_manager, "detect_runtime_environment", lambda: "vscode-server"
    )
    monkeypatch.setenv("PATH", "")

    asyncio.run(manager.find_latest_code(update_environment=True))

    expected = str(root / "bin" / "commit2" / "bin" / "remote-cli")
    assert str(manager.code_path) == expected
    assert os.environ["PATH"].split(os.pathsep)[0] == expected


def test_find_latest_code_reorders_when_path_has_duplicates(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path / "vscode"
    _write_code_launcher(
        root / "bin" / "commit9" / "bin" / "remote-cli" / "code", "9.0.0", "commit9"
    )

    expected = str(root / "bin" / "commit9" / "bin" / "remote-cli")
    monkeypatch.setenv(
        "PATH",
        f"alpha{os.pathsep}{expected}{os.pathsep}beta{os.pathsep}{expected}",
    )
    monkeypatch.setattr(code_manager, "vscode_root", root)
    monkeypatch.setattr(
        code_manager, "detect_runtime_environment", lambda: "vscode-server"
    )

    manager = CodeManager.__new__(CodeManager)
    manager.socket_path = None
    manager.code_path = None
    asyncio.run(manager.find_latest_code(update_environment=True))

    path_items = os.environ["PATH"].split(os.pathsep)
    assert path_items[0] == expected
    assert path_items.count(expected) == 1


def test_find_latest_code_without_environment_update_keeps_path(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path / "vscode"
    _write_code_launcher(
        root / "bin" / "commit3" / "bin" / "remote-cli" / "code", "3.0.0", "commit3"
    )

    manager = CodeManager.__new__(CodeManager)
    manager.socket_path = None
    manager.code_path = None

    monkeypatch.setattr(code_manager, "vscode_root", root)
    monkeypatch.setattr(
        code_manager, "detect_runtime_environment", lambda: "vscode-server"
    )
    original_path = f"alpha{os.pathsep}beta"
    monkeypatch.setenv("PATH", original_path)

    asyncio.run(manager.find_latest_code(update_environment=False))

    assert manager.code_path == root / "bin" / "commit3" / "bin" / "remote-cli"
    assert os.environ["PATH"] == original_path


def test_find_latest_code_handles_missing_or_invalid_versions(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture
) -> None:
    root = tmp_path / "vscode"
    manager = CodeManager.__new__(CodeManager)
    manager.socket_path = None
    manager.code_path = None

    monkeypatch.setattr(code_manager, "vscode_root", root)
    monkeypatch.setattr(
        code_manager, "detect_runtime_environment", lambda: "vscode-server"
    )
    monkeypatch.setattr(code_manager.shutil, "which", lambda _name: None)

    with caplog.at_level("WARNING"):
        asyncio.run(manager.find_latest_code(update_environment=True))

    assert "No VSCode CLI executable found" in caplog.text

    _write_code_launcher(
        root / "bin" / "commit-a" / "bin" / "remote-cli" / "code", "1.0.0", "commit-a"
    )
    original_max = builtins.max

    def _max(values, *args, **kwargs):
        if not args and not kwargs:
            return "missing"
        return original_max(values, *args, **kwargs)

    monkeypatch.setattr(builtins, "max", _max)

    with caplog.at_level("WARNING"):
        asyncio.run(manager.find_latest_code(update_environment=True))

    assert "Unable to determine latest VSCode version" in caplog.text


def test_find_latest_code_ignores_launcher_missing_commit(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture
) -> None:
    root = tmp_path / "vscode"
    script = root / "bin" / "commit-x" / "bin" / "remote-cli" / "code"
    script.parent.mkdir(parents=True, exist_ok=True)
    script.write_text("VERSION=1.0.0\n")

    manager = CodeManager.__new__(CodeManager)
    manager.socket_path = None
    manager.code_path = None

    monkeypatch.setattr(code_manager, "vscode_root", root)
    monkeypatch.setattr(
        code_manager, "detect_runtime_environment", lambda: "vscode-server"
    )
    monkeypatch.setattr(code_manager.shutil, "which", lambda _name: None)

    with caplog.at_level("WARNING"):
        asyncio.run(manager.find_latest_code(update_environment=False))

    assert "No VSCode CLI executable found" in caplog.text


def test_find_latest_code_falls_back_to_local_cli(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    manager = CodeManager.__new__(CodeManager)
    manager.socket_path = None
    manager.code_path = None
    temp_dir = Path(tempfile.gettempdir())

    monkeypatch.setattr(code_manager, "detect_runtime_environment", lambda: "local")
    monkeypatch.setattr(
        code_manager.shutil,
        "which",
        lambda name: str(temp_dir / "code") if name == "code" else None,
    )
    monkeypatch.setenv("PATH", "")

    asyncio.run(manager.find_latest_code(update_environment=True))

    expected_path = temp_dir.resolve()
    assert manager.code_path == expected_path
    assert Path(os.environ["PATH"].split(os.pathsep)[0]).resolve() == expected_path


@pytest.mark.parametrize(
    ("error", "expected"),
    [
        (BlockingIOError(), False),
        (ConnectionResetError(), True),
        (ConnectionRefusedError(), True),
        (RuntimeError("unexpected"), False),
    ],
)
def test_is_socket_closed_branches(
    monkeypatch: pytest.MonkeyPatch,
    error: Exception,
    expected: bool,
) -> None:
    monkeypatch.setattr(socket, "AF_UNIX", getattr(socket, "AF_UNIX", 1), raising=False)
    monkeypatch.setattr(
        socket,
        "MSG_DONTWAIT",
        getattr(socket, "MSG_DONTWAIT", 0x40),
        raising=False,
    )
    monkeypatch.setattr(
        socket,
        "MSG_PEEK",
        getattr(socket, "MSG_PEEK", 0x02),
        raising=False,
    )

    class _FakeSocket:
        def connect(self, _value: str) -> None:
            return None

        def recv(self, _size: int, _flags: int) -> None:
            raise error

    monkeypatch.setattr(socket, "socket", lambda *_args, **_kwargs: _FakeSocket())

    assert (
        CodeManager.is_socket_closed(Path(tempfile.gettempdir()) / "fake.sock")
        is expected
    )


def test_is_socket_closed_returns_false_when_recv_succeeds(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(socket, "AF_UNIX", getattr(socket, "AF_UNIX", 1), raising=False)
    monkeypatch.setattr(
        socket,
        "MSG_DONTWAIT",
        getattr(socket, "MSG_DONTWAIT", 0x40),
        raising=False,
    )
    monkeypatch.setattr(
        socket,
        "MSG_PEEK",
        getattr(socket, "MSG_PEEK", 0x02),
        raising=False,
    )

    class _FakeSocket:
        def connect(self, _value: str) -> None:
            return None

        def recv(self, _size: int, _flags: int) -> bytes:
            return b"ok"

    monkeypatch.setattr(socket, "socket", lambda *_args, **_kwargs: _FakeSocket())

    assert (
        CodeManager.is_socket_closed(Path(tempfile.gettempdir()) / "fake.sock") is False
    )


def test_is_socket_closed_returns_false_when_unix_socket_capabilities_missing(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.delattr(socket, "AF_UNIX", raising=False)

    assert (
        CodeManager.is_socket_closed(Path(tempfile.gettempdir()) / "fake.sock") is False
    )


def test_is_socket_closed_closes_socket(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(socket, "AF_UNIX", getattr(socket, "AF_UNIX", 1), raising=False)
    monkeypatch.setattr(
        socket,
        "MSG_DONTWAIT",
        getattr(socket, "MSG_DONTWAIT", 0x40),
        raising=False,
    )
    monkeypatch.setattr(
        socket,
        "MSG_PEEK",
        getattr(socket, "MSG_PEEK", 0x02),
        raising=False,
    )

    class _FakeSocket:
        def __init__(self) -> None:
            self.closed = False

        def connect(self, _value: str) -> None:
            return None

        def recv(self, _size: int, _flags: int) -> None:
            raise ConnectionRefusedError()

        def close(self) -> None:
            self.closed = True

    fake_socket = _FakeSocket()
    monkeypatch.setattr(socket, "socket", lambda *_args, **_kwargs: fake_socket)

    assert (
        CodeManager.is_socket_closed(Path(tempfile.gettempdir()) / "fake.sock") is True
    )
    assert fake_socket.closed is True


def test_is_socket_closed_handles_non_callable_close(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(socket, "AF_UNIX", getattr(socket, "AF_UNIX", 1), raising=False)
    monkeypatch.setattr(
        socket,
        "MSG_DONTWAIT",
        getattr(socket, "MSG_DONTWAIT", 0x40),
        raising=False,
    )
    monkeypatch.setattr(
        socket,
        "MSG_PEEK",
        getattr(socket, "MSG_PEEK", 0x02),
        raising=False,
    )

    class _FakeSocket:
        close = None

        def connect(self, _value: str) -> None:
            return None

        def recv(self, _size: int, _flags: int) -> None:
            raise ConnectionRefusedError()

    monkeypatch.setattr(socket, "socket", lambda *_args, **_kwargs: _FakeSocket())

    assert (
        CodeManager.is_socket_closed(Path(tempfile.gettempdir()) / "fake.sock") is True
    )


def test_is_socket_closed_handles_socket_creation_failure(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(socket, "AF_UNIX", getattr(socket, "AF_UNIX", 1), raising=False)
    monkeypatch.setattr(
        socket,
        "MSG_DONTWAIT",
        getattr(socket, "MSG_DONTWAIT", 0x40),
        raising=False,
    )
    monkeypatch.setattr(
        socket,
        "MSG_PEEK",
        getattr(socket, "MSG_PEEK", 0x02),
        raising=False,
    )

    def _raise_socket(*_args, **_kwargs):
        raise RuntimeError("boom")

    monkeypatch.setattr(socket, "socket", _raise_socket)

    assert (
        CodeManager.is_socket_closed(Path(tempfile.gettempdir()) / "fake.sock") is False
    )


def test_async_code_manager_methods_delegate_to_private_sync_methods() -> None:
    manager = CodeManager.__new__(CodeManager)
    calls: list[str] = []

    manager._find_socket_sync = lambda update_environment=False: calls.append(
        f"socket:{update_environment}"
    )
    manager._find_latest_code_sync = lambda update_environment=False: calls.append(
        f"code:{update_environment}"
    )

    asyncio.run(manager.find_socket(update_environment=True))
    asyncio.run(manager.find_latest_code(update_environment=False))
    asyncio.run(manager.initialize())

    assert calls == ["socket:True", "code:False", "socket:True", "code:True"]



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class ExtensionSpec:
    extension_id: str
    pinned_version: str = ""


@dataclass(frozen=True)
class ResolvedExtensionRequest:
    extension_id: str
    pinned_version: str = ""



from __future__ import annotations

import asyncio
from types import SimpleNamespace

from uvscem.api_client import CodeAPIManager


def test_init_creates_session_with_http_adapters() -> None:
    manager = CodeAPIManager()

    assert "https://" in manager.session.adapters
    assert "http://" in manager.session.adapters
    retry = manager.session.adapters["https://"].max_retries
    assert "POST" in retry.allowed_methods


def test_get_vscode_extension_handles_pagination_and_flags() -> None:
    manager = CodeAPIManager.__new__(CodeAPIManager)
    calls: list[dict] = []

    class _Response:
        def __init__(self, extensions: list[dict]) -> None:
            self._extensions = extensions

        def raise_for_status(self) -> None:
            return None

        def json(self) -> dict:
            return {"results": [{"extensions": self._extensions}]}

    def _post(url: str, json: dict, headers: dict, timeout: int) -> _Response:
        calls.append({"url": url, "json": json, "headers": headers})
        assert timeout == 30
        if len(calls) == 1:
            return _Response([{"id": "one"}, {"id": "two"}])
        return _Response([{"id": "three"}])

    manager.session = SimpleNamespace(post=_post)

    result = asyncio.run(
        manager.get_vscode_extension(
            extension_id="publisher.name",
            page_size=2,
            include_category_and_tags=False,
            include_asset_uri=False,
            include_latest_version_only=True,
            unpublished=True,
            include_name_conflict_info=True,
        )
    )

    assert [item["id"] for item in result] == ["one", "two", "three"]
    assert len(calls) == 2
    assert calls[0]["headers"]["Accept"].endswith("api-version=7.2-preview.1")
    assert calls[0]["json"]["filters"][0]["criteria"][1]["value"] == "publisher.name"
    assert calls[0]["json"]["flags"] == (
        0x1 | 0x2 | 0x8 | 0x10 | 0x40 | 0x100 | 0x200 | 0x1000 | 0x8000
    )


def test_get_vscode_extension_supports_minimal_flag_set() -> None:
    manager = CodeAPIManager.__new__(CodeAPIManager)
    calls: list[dict] = []

    class _Response:
        def raise_for_status(self) -> None:
            return None

        def json(self) -> dict:
            return {"results": [{"extensions": []}]}

    manager.session = SimpleNamespace(
        post=lambda url, json, headers, timeout: (
            calls.append({"url": url, "json": json, "headers": headers}) or _Response()
        )
    )

    result = asyncio.run(
        manager.get_vscode_extension(
            extension_id="publisher.name",
            max_page=1,
            page_size=10,
            include_versions=False,
            include_files=False,
            include_category_and_tags=False,
            include_shared_accounts=False,
            include_version_properties=False,
            exclude_non_validated=False,
            include_installation_targets=False,
            include_asset_uri=False,
            include_statistics=False,
            include_latest_version_only=False,
            unpublished=False,
            include_name_conflict_info=False,
            api_version="1.0-preview",
        )
    )

    assert result == []
    assert calls[0]["json"]["flags"] == 0
    assert calls[0]["headers"]["Accept"].endswith("api-version=1.0-preview")


def test_get_vscode_extension_covers_additional_flag_paths() -> None:
    manager = CodeAPIManager.__new__(CodeAPIManager)

    class _Response:
        def raise_for_status(self) -> None:
            return None

        def json(self) -> dict:
            return {"results": [{"extensions": []}]}

    manager.session = SimpleNamespace(post=lambda *_args, **_kwargs: _Response())

    result = asyncio.run(
        manager.get_vscode_extension(
            max_page=0,
            include_versions=False,
            include_files=False,
            include_category_and_tags=True,
            include_shared_accounts=False,
            include_version_properties=False,
            exclude_non_validated=True,
            include_installation_targets=False,
            include_asset_uri=True,
            include_statistics=False,
            include_latest_version_only=False,
            unpublished=False,
            include_name_conflict_info=False,
        )
    )

    assert result == []


def _build_extension_payload() -> dict:
    return {
        "extensionId": "ext-guid",
        "extensionName": "sample-extension",
        "statistics": [{"statisticName": "install", "value": 42}],
        "publisher": {
            "publisherName": "sample-publisher",
            "displayName": "Sample Publisher",
            "publisherId": "publisher-guid",
        },
        "versions": [
            {
                "version": "2.0.0-pre",
                "properties": [
                    {
                        "key": "Microsoft.VisualStudio.Code.PreRelease",
                        "value": True,
                    },
                    {
                        "key": "Microsoft.VisualStudio.Code.ExtensionDependencies",
                        "value": "dep.one,dep.two",
                    },
                    {
                        "key": "Microsoft.VisualStudio.Code.ExtensionPack",
                        "value": "pack.one",
                    },
                ],
                "files": [
                    {
                        "assetType": "Microsoft.VisualStudio.Services.VSIXPackage",
                        "source": "https://download/pre.vsix",
                    },
                    {
                        "assetType": "Microsoft.VisualStudio.Services.VsixSignature",
                        "source": "https://download/pre.sig",
                    },
                ],
            },
            {
                "version": "1.9.0",
                "properties": [
                    {
                        "key": "Microsoft.VisualStudio.Code.PreRelease",
                        "value": False,
                    },
                    {
                        "key": "Microsoft.VisualStudio.Code.ExtensionDependencies",
                        "value": "dep.one,dep.two",
                    },
                    {
                        "key": "Microsoft.VisualStudio.Code.ExtensionPack",
                        "value": "pack.one",
                    },
                ],
                "files": [
                    {
                        "assetType": "Microsoft.VisualStudio.Services.VSIXPackage",
                        "source": "https://download/stable.vsix",
                    },
                    {
                        "assetType": "Microsoft.VisualStudio.Services.VsixSignature",
                        "source": "https://download/stable.sig",
                    },
                ],
            },
        ],
    }


def test_get_extension_metadata_skips_prerelease_by_default() -> None:
    manager = CodeAPIManager.__new__(CodeAPIManager)
    manager._get_vscode_extension_sync = lambda **_kwargs: [_build_extension_payload()]

    result = asyncio.run(manager.get_extension_metadata("publisher.name"))
    versions = result["publisher.name"]

    assert len(versions) == 1
    metadata = versions[0]
    assert metadata["version"] == "1.9.0"
    assert metadata["dependencies"] == ["dep.one", "dep.two"]
    assert metadata["extension_pack"] == ["pack.one"]
    assert metadata["url"] == "https://download/stable.vsix"
    assert metadata["signature"] == "https://download/stable.sig"
    assert metadata["statistics"]["install"] == 42
    assert metadata["installation_metadata"]["identifier"]["id"] == "publisher.name"
    assert metadata["installation_metadata"]["metadata"]["isPreReleaseVersion"] is False


def test_get_extension_metadata_can_include_prerelease_versions() -> None:
    manager = CodeAPIManager.__new__(CodeAPIManager)
    manager._get_vscode_extension_sync = lambda **_kwargs: [_build_extension_payload()]

    result = asyncio.run(
        manager.get_extension_metadata(
            "publisher.name", include_latest_stable_version_only=False
        )
    )
    versions = result["publisher.name"]

    assert [item["version"] for item in versions] == ["2.0.0-pre", "1.9.0"]
    assert (
        versions[0]["installation_metadata"]["metadata"]["isPreReleaseVersion"] is True
    )


def test_get_extension_metadata_handles_extension_without_versions() -> None:
    manager = CodeAPIManager.__new__(CodeAPIManager)
    manager._get_vscode_extension_sync = lambda **_kwargs: [
        {
            "extensionId": "id",
            "extensionName": "name",
            "statistics": [],
            "publisher": {"publisherName": "publisher"},
            "versions": [],
        }
    ]

    result = asyncio.run(manager.get_extension_metadata("publisher.name"))

    assert result == {"publisher.name": []}


def test_get_extension_metadata_can_select_requested_version() -> None:
    manager = CodeAPIManager.__new__(CodeAPIManager)
    manager._get_vscode_extension_sync = lambda **_kwargs: [_build_extension_payload()]

    result = asyncio.run(
        manager.get_extension_metadata(
            "publisher.name",
            include_latest_stable_version_only=False,
            requested_version="2.0.0-pre",
        )
    )
    versions = result["publisher.name"]

    assert [item["version"] for item in versions] == ["2.0.0-pre"]


def test_get_extension_metadata_requested_version_returns_empty_when_missing() -> None:
    manager = CodeAPIManager.__new__(CodeAPIManager)
    manager._get_vscode_extension_sync = lambda **_kwargs: [_build_extension_payload()]

    result = asyncio.run(
        manager.get_extension_metadata(
            "publisher.name",
            include_latest_stable_version_only=False,
            requested_version="9.9.9",
        )
    )

    assert result == {"publisher.name": []}


def test_async_api_methods_use_private_sync_helpers() -> None:
    manager = CodeAPIManager.__new__(CodeAPIManager)
    manager._get_vscode_extension_sync = lambda **_kwargs: [
        {"id": "one"},
        {"id": "two"},
    ]
    manager._get_extension_metadata_sync = lambda *args, **kwargs: {
        "publisher.name": []
    }

    extensions = asyncio.run(
        manager.get_vscode_extension(extension_id="publisher.name")
    )
    metadata = asyncio.run(manager.get_extension_metadata("publisher.name"))

    assert extensions == [{"id": "one"}, {"id": "two"}]
    assert metadata == {"publisher.name": []}



from __future__ import annotations

import datetime
import hashlib
from pathlib import Path
from typing import Any


def sha256_file(path: Path) -> str:
    return hashlib.sha256(path.read_bytes()).hexdigest()


def resolve_vsce_sign_targets(
    value: str,
    current_target: str,
    supported_targets: tuple[str, ...],
) -> list[str]:
    if value == "current":
        return [current_target]
    if value == "all":
        return list(supported_targets)
    result = [item.strip() for item in value.split(",") if item.strip()]
    if not result:
        raise ValueError("No valid vsce-sign targets were provided")
    return result


def build_extension_manifest_entry(
    extension_id: str,
    metadata: dict[str, Any],
    artifacts_dir: Path,
) -> dict[str, Any]:
    version = str(metadata.get("version", ""))
    filename = f"{extension_id}-{version}.vsix"
    signature_filename = f"{extension_id}-{version}.sigzip"
    return {
        "id": extension_id,
        "version": version,
        "filename": filename,
        "signature_filename": signature_filename,
        "vsix_sha256": sha256_file(artifacts_dir.joinpath(filename)),
        "signature_sha256": sha256_file(artifacts_dir.joinpath(signature_filename)),
        "installation_metadata": metadata.get("installation_metadata", {}),
        "dependencies": list(metadata.get("dependencies", [])),
        "extension_pack": list(metadata.get("extension_pack", [])),
    }


def build_bundle_manifest(
    config_name: str,
    ordered_extensions: list[str],
    extension_entries: list[dict[str, Any]],
    vsce_sign_version: str,
    vsce_sign_binaries: list[dict[str, str]],
) -> dict[str, Any]:
    return {
        "schema_version": 1,
        "created_at": datetime.datetime.now(datetime.timezone.utc).isoformat(),
        "config_name": config_name,
        "ordered_extensions": ordered_extensions,
        "extensions": extension_entries,
        "vsce_sign": {
            "version": vsce_sign_version,
            "binaries": vsce_sign_binaries,
        },
    }


def resolve_bundled_vsce_sign_binary(
    bundle_dir: Path,
    vsce_sign_info: dict[str, Any],
    current_target: str,
) -> tuple[Path, str]:
    binaries = list(vsce_sign_info.get("binaries", []))
    if binaries:
        selected = None
        for item in binaries:
            if str(item.get("target", "")) == current_target:
                selected = item
                break
        if selected is None and len(binaries) == 1:
            selected = binaries[0]
        if selected is None:
            raise ValueError(
                f"No bundled vsce-sign binary found for target {current_target}"
            )

        binary_rel = str(selected.get("binary", ""))
        return bundle_dir.joinpath(binary_rel), str(selected.get("sha256", ""))

    binary_rel = str(vsce_sign_info.get("binary", ""))
    return bundle_dir.joinpath(binary_rel), ""



from __future__ import annotations

DEFAULT_USER_AGENT = (
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
    "AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.3.1 Safari/605.1.15"
)

MAX_INSTALL_RETRIES = 3

HTTP_REQUEST_TIMEOUT_SECONDS = 30
HTTP_STREAM_CONNECT_TIMEOUT_SECONDS = 10
HTTP_STREAM_READ_TIMEOUT_SECONDS = 120

HTTP_RETRY_TOTAL = 3
HTTP_RETRY_BACKOFF_FACTOR = 1
HTTP_RETRY_STATUS_FORCELIST = [429, 500, 502, 503, 504]
HTTP_RETRY_ALLOWED_METHODS = ["HEAD", "GET", "OPTIONS", "POST"]



#! /bin/env python3
from __future__ import annotations

import asyncio
import logging
from pathlib import Path
from typing import Any

# for parsing devcontainer.json (if it includes comments etc.)
import requests
from requests.adapters import HTTPAdapter, Retry

from uvscem.internal_config import (
    DEFAULT_USER_AGENT,
    HTTP_REQUEST_TIMEOUT_SECONDS,
    HTTP_RETRY_ALLOWED_METHODS,
    HTTP_RETRY_BACKOFF_FACTOR,
    HTTP_RETRY_STATUS_FORCELIST,
    HTTP_RETRY_TOTAL,
    MAX_INSTALL_RETRIES,
)
from uvscem.marketplace import (
    build_extension_query_body,
    build_query_flags,
    shape_extension_metadata_versions,
)
from uvscem.vscode_paths import resolve_vscode_root

__author__ = "Arne Fahrenwalde <arne@fahrenwal.de>"


# attempt to install an extension a maximum of three times
max_retries = MAX_INSTALL_RETRIES
user_agent: str = DEFAULT_USER_AGENT
# VSCode extension installation directory
vscode_root: Path = resolve_vscode_root()
logger: logging.Logger = logging.getLogger(__name__)


class CodeAPIManager(object):
    """Directly obtain all relevant information from the VSCode Marketplace API."""

    session: requests.Session

    def _get_vscode_extension_sync(
        self,
        extension_id: str = "",
        max_page: int = 100,
        page_size: int = 10,
        include_versions: bool = True,
        include_files: bool = True,
        include_category_and_tags: bool = True,
        include_shared_accounts: bool = True,
        include_version_properties: bool = True,
        exclude_non_validated: bool = False,
        include_installation_targets: bool = True,
        include_asset_uri: bool = True,
        include_statistics: bool = True,
        include_latest_version_only: bool = False,
        unpublished: bool = False,
        include_name_conflict_info: bool = True,
        api_version="7.2-preview.1",
    ) -> Any:
        headers = {
            "Accept": f"application/json; charset=utf-8; api-version={api_version}"
        }

        flags = build_query_flags(
            include_versions=include_versions,
            include_files=include_files,
            include_category_and_tags=include_category_and_tags,
            include_shared_accounts=include_shared_accounts,
            include_version_properties=include_version_properties,
            exclude_non_validated=exclude_non_validated,
            include_installation_targets=include_installation_targets,
            include_asset_uri=include_asset_uri,
            include_statistics=include_statistics,
            include_latest_version_only=include_latest_version_only,
            unpublished=unpublished,
            include_name_conflict_info=include_name_conflict_info,
        )

        for page in range(1, max_page + 1):
            """
            filterType:
                Tag = 1,
                ExtensionId = 4,
                Category = 5,
                ExtensionName = 7,
                Target = 8,
                Featured = 9,
                SearchText = 10,
                ExcludeWithFlags = 12

            SortBy:
                NoneOrRelevance = 0,
                LastUpdatedDate = 1,
                Title = 2,
                PublisherName = 3,
                InstallCount = 4,
                PublishedDate = 10,
                AverageRating = 6,
                WeightedRating = 12

            SortOrder:
                Default = 0,
                Ascending = 1,
                Descending = 2
            """
            body = build_extension_query_body(
                extension_id=extension_id,
                page_number=page,
                page_size=page_size,
                flags=flags,
            )

            r = self.session.post(
                "https://marketplace.visualstudio.com/_apis/public/gallery/extensionquery",
                json=body,
                headers=headers,
                timeout=HTTP_REQUEST_TIMEOUT_SECONDS,
            )
            r.raise_for_status()
            response = r.json()

            extensions = response["results"][0]["extensions"]
            for extension in extensions:
                yield extension

            if len(extensions) != page_size:
                break

    def _get_extension_metadata_sync(
        self,
        extension_id: str,
        include_latest_version_only: bool = False,
        include_latest_stable_version_only: bool = True,
        requested_version: str = "",
    ) -> dict[str, list[dict[str, Any]]]:
        extensions = {}
        logger.info(f"Obtaining metadata for {extension_id}")
        for extension in self._get_vscode_extension_sync(
            extension_id=extension_id,
            include_latest_version_only=include_latest_version_only,
        ):
            result = shape_extension_metadata_versions(
                extension=extension,
                extension_id=extension_id,
                include_latest_stable_version_only=include_latest_stable_version_only,
                requested_version=requested_version,
                vscode_root=vscode_root,
            )
            for item in result:
                logger.debug(f"- {item.get('url', '')}")
            extensions[extension_id] = result
        return extensions

    async def get_vscode_extension(
        self,
        extension_id: str = "",
        max_page: int = 100,
        page_size: int = 10,
        include_versions: bool = True,
        include_files: bool = True,
        include_category_and_tags: bool = True,
        include_shared_accounts: bool = True,
        include_version_properties: bool = True,
        exclude_non_validated: bool = False,
        include_installation_targets: bool = True,
        include_asset_uri: bool = True,
        include_statistics: bool = True,
        include_latest_version_only: bool = False,
        unpublished: bool = False,
        include_name_conflict_info: bool = True,
        api_version: str = "7.2-preview.1",
    ) -> list[dict[str, Any]]:
        """Asynchronously fetch extension search results."""
        return await asyncio.to_thread(
            lambda: list(
                self._get_vscode_extension_sync(
                    extension_id=extension_id,
                    max_page=max_page,
                    page_size=page_size,
                    include_versions=include_versions,
                    include_files=include_files,
                    include_category_and_tags=include_category_and_tags,
                    include_shared_accounts=include_shared_accounts,
                    include_version_properties=include_version_properties,
                    exclude_non_validated=exclude_non_validated,
                    include_installation_targets=include_installation_targets,
                    include_asset_uri=include_asset_uri,
                    include_statistics=include_statistics,
                    include_latest_version_only=include_latest_version_only,
                    unpublished=unpublished,
                    include_name_conflict_info=include_name_conflict_info,
                    api_version=api_version,
                )
            )
        )

    async def get_extension_metadata(
        self,
        extension_id: str,
        include_latest_version_only: bool = False,
        include_latest_stable_version_only: bool = True,
        requested_version: str = "",
    ) -> dict[str, list[dict[str, Any]]]:
        """Asynchronously fetch metadata for one extension."""
        return await asyncio.to_thread(
            self._get_extension_metadata_sync,
            extension_id,
            include_latest_version_only,
            include_latest_stable_version_only,
            requested_version,
        )

    def __init__(self):
        retry_strategy = Retry(
            total=HTTP_RETRY_TOTAL,
            backoff_factor=HTTP_RETRY_BACKOFF_FACTOR,
            status_forcelist=HTTP_RETRY_STATUS_FORCELIST,
            allowed_methods=HTTP_RETRY_ALLOWED_METHODS,
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        self.session = requests.Session()
        self.session.mount("https://", adapter)
        self.session.mount("http://", adapter)



from __future__ import annotations

import asyncio
import json
import os
import shlex
import shutil
import subprocess
from contextlib import contextmanager
from pathlib import Path

import pytest

import uvscem.code_manager as code_manager_module
import uvscem.extension_manager as extension_manager_module
from uvscem.extension_manager import CodeExtensionManager
from uvscem.vsce_sign_bootstrap import provision_vsce_sign_binary_for_run

pytestmark = pytest.mark.slow


ASSETS_DIR = Path(__file__).parent / "assets"


def _artifact_cache_dir(tmp_path: Path, name: str) -> Path:
    cache_root = os.environ.get("UVSCEM_INTEGRATION_ARTIFACT_CACHE", "").strip()
    if cache_root:
        base_path = Path(cache_root).expanduser()
    else:
        project_root = Path(__file__).resolve().parents[2]
        base_path = project_root / ".tmp" / "uvscem-test-artifacts"

    try:
        base_path.mkdir(parents=True, exist_ok=True)
    except OSError:
        base_path = Path.home() / ".cache" / "uvscem-test-artifacts"
        try:
            base_path.mkdir(parents=True, exist_ok=True)
        except OSError:
            base_path = tmp_path / "cache"
    target_path = base_path / name
    target_path.mkdir(parents=True, exist_ok=True)
    return target_path


def _clear_cached_extension_artifacts(tmp_path: Path, extension_id: str) -> None:
    cache_dir = _artifact_cache_dir(tmp_path, "downloads")
    for file_path in cache_dir.glob(f"{extension_id}-*.vsix"):
        file_path.unlink(missing_ok=True)
    for file_path in cache_dir.glob(f"{extension_id}-*.sigzip"):
        file_path.unlink(missing_ok=True)


@pytest.fixture(autouse=True)
def _mock_vscode_socket_paths_for_ci(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path_factory: pytest.TempPathFactory,
) -> None:
    if os.environ.get("GITHUB_ACTIONS", "").lower() != "true":
        return

    fake_socket_path = tmp_path_factory.mktemp("vscode-ipc") / "vscode-ipc-ci.sock"

    async def _fake_initialize(self) -> None:
        self.socket_path = fake_socket_path
        monkeypatch.setenv("VSCODE_IPC_HOOK_CLI", str(fake_socket_path))

    async def _fake_find_socket(self, update_environment: bool = False) -> None:
        self.socket_path = fake_socket_path
        if update_environment:
            monkeypatch.setenv("VSCODE_IPC_HOOK_CLI", str(fake_socket_path))

    monkeypatch.setattr(code_manager_module.CodeManager, "initialize", _fake_initialize)
    monkeypatch.setattr(
        code_manager_module.CodeManager, "find_socket", _fake_find_socket
    )


def _load_extensions(config_path: Path) -> list[str]:
    data = json.loads(config_path.read_text())
    return data.get("customizations", {}).get("vscode", {}).get("extensions", [])


def _installed_extensions(code_binary: str) -> set[str]:
    result = subprocess.run(
        [code_binary, "--list-extensions"],
        capture_output=True,
        check=True,
        text=True,
    )
    return {line.strip() for line in result.stdout.splitlines() if line.strip()}


def _uninstall_extension(code_binary: str, extension_id: str) -> None:
    subprocess.run(
        [code_binary, "--uninstall-extension", extension_id],
        capture_output=True,
        check=False,
        text=True,
    )


def _isolated_code_binary(real_code_binary: str, tmp_path: Path) -> str:
    extensions_dir = tmp_path / "extensions-dir"
    user_data_dir = tmp_path / "user-data-dir"
    extensions_dir.mkdir(parents=True, exist_ok=True)
    user_data_dir.mkdir(parents=True, exist_ok=True)

    wrapper = tmp_path / "code-isolated"
    wrapper.write_text(
        "#!/usr/bin/env sh\n"
        f"exec {shlex.quote(real_code_binary)} "
        f"--extensions-dir {shlex.quote(str(extensions_dir))} "
        f'--user-data-dir {shlex.quote(str(user_data_dir))} "$@"\n',
        encoding="utf-8",
    )
    wrapper.chmod(0o755)
    return str(wrapper)


@contextmanager
def _misconfigured_proxy(monkeypatch: pytest.MonkeyPatch):
    proxy_url = "http://127.0.0.1:9"
    monkeypatch.setenv("HTTP_PROXY", proxy_url)
    monkeypatch.setenv("HTTPS_PROXY", proxy_url)
    monkeypatch.setenv("http_proxy", proxy_url)
    monkeypatch.setenv("https_proxy", proxy_url)
    monkeypatch.setenv("NO_PROXY", "")
    monkeypatch.setenv("no_proxy", "")
    yield


@pytest.mark.slow
def test_integration_fetch_install_uninstall_with_verification(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    code_binary = shutil.which("code")
    isolated_code_binary = (
        _isolated_code_binary(code_binary, tmp_path) if code_binary else "code"
    )
    sandbox_vscode_root = tmp_path / "vscode-server"
    sandbox_extensions = sandbox_vscode_root / "extensions"
    sandbox_extensions.mkdir(parents=True, exist_ok=True)
    (sandbox_extensions / "extensions.json").write_text("[]", encoding="utf-8")

    monkeypatch.setattr(extension_manager_module, "vscode_root", sandbox_vscode_root)

    config_path = ASSETS_DIR / "test_extensions.json"

    manager = CodeExtensionManager(
        config_name=str(config_path),
        code_path=isolated_code_binary,
        target_directory=str(_artifact_cache_dir(tmp_path, "downloads")),
    )
    asyncio.run(manager.initialize())
    extensions = manager.extensions or _load_extensions(config_path)

    try:
        if not code_binary:
            pytest.skip("VS Code CLI not available for install/uninstall validation")
        asyncio.run(manager.install_async())

        for extension_id in extensions:
            extension_dir = sandbox_extensions / manager.get_dirname(extension_id)
            assert extension_dir.is_dir()

        installed_metadata = json.loads(
            (sandbox_extensions / "extensions.json").read_text(encoding="utf-8")
        )
        installed_ids = {
            str(entry.get("identifier", {}).get("id", ""))
            for entry in installed_metadata
            if isinstance(entry, dict)
        }
        for extension_id in extensions:
            assert extension_id in installed_ids
    finally:
        for extension_id in extensions:
            extension_dir = sandbox_extensions / manager.get_dirname(extension_id)
            if extension_dir.is_dir():
                shutil.rmtree(extension_dir)
        (sandbox_extensions / "extensions.json").write_text("[]", encoding="utf-8")

    installed_after_cleanup = json.loads(
        (sandbox_extensions / "extensions.json").read_text(encoding="utf-8")
    )
    installed_after_cleanup_ids = {
        str(entry.get("identifier", {}).get("id", ""))
        for entry in installed_after_cleanup
        if isinstance(entry, dict)
    }
    for extension_id in extensions:
        assert extension_id not in installed_after_cleanup_ids
        extension_dir = sandbox_extensions / manager.get_dirname(extension_id)
        assert not extension_dir.exists()


@pytest.mark.slow
def test_integration_tampered_vsix_fails_signature_verification(tmp_path: Path) -> None:
    config_path = ASSETS_DIR / "test_extensions.json"
    extension_id = "dbaeumer.vscode-eslint"

    manager = CodeExtensionManager(
        config_name=str(config_path),
        code_path=shutil.which("code") or "code",
        target_directory=str(_artifact_cache_dir(tmp_path, "downloads")),
    )

    metadata_map = asyncio.run(manager.api_manager.get_extension_metadata(extension_id))
    versions = metadata_map.get(extension_id, [])
    if not versions:
        pytest.skip("Could not resolve extension metadata for slow integration test")
    manager.extension_metadata[extension_id] = versions[0]

    vsix_path = asyncio.run(manager.download_extension(extension_id))
    signature_path = asyncio.run(manager.download_signature_archive(extension_id))

    payload = bytearray(vsix_path.read_bytes())
    if not payload:
        pytest.skip("Downloaded VSIX is empty")
    payload[len(payload) // 2] ^= 0x01
    tampered_vsix_path = tmp_path / "tampered.vsix"
    tampered_vsix_path.write_bytes(bytes(payload))

    with provision_vsce_sign_binary_for_run(
        install_dir=_artifact_cache_dir(tmp_path, "vsce-sign-bin"),
        verify_existing_checksum=False,
    ) as binary:
        manager.vsce_sign_binary = binary
        try:
            with pytest.raises(subprocess.CalledProcessError):
                asyncio.run(
                    manager.verify_extension_signature(
                        tampered_vsix_path, signature_path
                    )
                )
        finally:
            manager.vsce_sign_binary = None


@pytest.mark.slow
@pytest.mark.parametrize(
    "extension_id",
    ["dbaeumer.vscode-eslint", "cristianvasquez1312.hadar-vscode"],
    ids=["extension", "extension-pack"],
)
@pytest.mark.parametrize("pinned", [False, True], ids=["unpinned", "pinned-older"])
def test_integration_offline_bundle_import_ci_compatible_with_misconfigured_proxy(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    extension_id: str,
    pinned: bool,
) -> None:
    _clear_cached_extension_artifacts(tmp_path, extension_id)

    if pinned:
        metadata_map = asyncio.run(
            CodeExtensionManager(
                config_name=str(ASSETS_DIR / "test_extensions.json"),
                code_path=shutil.which("code") or "code",
                target_directory=str(_artifact_cache_dir(tmp_path, "downloads")),
            ).api_manager.get_extension_metadata(
                extension_id,
                include_latest_stable_version_only=False,
            )
        )
        versions = metadata_map.get(extension_id, [])
        latest_version = str(versions[0].get("version", "")) if versions else ""
        pinned_metadata = next(
            (
                version
                for version in versions[1:]
                if str(version.get("version", ""))
                and str(version.get("signature", ""))
                and str(version.get("version", "")) != latest_version
            ),
            None,
        )
        if pinned_metadata is None:
            pinned_metadata = next(
                (
                    version
                    for version in versions
                    if str(version.get("version", ""))
                    and str(version.get("signature", ""))
                ),
                None,
            )
        if pinned_metadata is None:
            pytest.skip("Could not resolve older extension version for pinned test")
        resolved_version = str(pinned_metadata.get("version", ""))
        if not resolved_version:
            pytest.skip("Could not resolve extension version for pinned test")
        extension_spec = f"{extension_id}@{resolved_version}"
    else:
        resolved_version = ""
        extension_spec = extension_id

    config_path = tmp_path / (
        "test_extensions_pinned.json" if pinned else "test_extensions_unpinned.json"
    )
    config_path.write_text(
        json.dumps(
            {
                "customizations": {
                    "vscode": {
                        "extensions": [extension_spec],
                    }
                }
            }
        ),
        encoding="utf-8",
    )

    bundle_path = tmp_path / "bundle"
    code_binary = shutil.which("code")
    sandbox_vscode_root = tmp_path / "vscode-server"
    sandbox_extensions = sandbox_vscode_root / "extensions"
    sandbox_extensions.mkdir(parents=True, exist_ok=True)
    (sandbox_extensions / "extensions.json").write_text("[]", encoding="utf-8")

    monkeypatch.setattr(extension_manager_module, "vscode_root", sandbox_vscode_root)
    ordered_extensions: list[str] = []
    try:
        extension_manager_module.export_offline_bundle(
            config_name=str(config_path),
            bundle_path=str(bundle_path),
            target_path=str(_artifact_cache_dir(tmp_path, "downloads")),
            code_path=code_binary or "code",
        )

        manifest = json.loads(
            (bundle_path / "manifest.json").read_text(encoding="utf-8")
        )
        ordered_extensions = [
            str(resolved_extension_id)
            for resolved_extension_id in manifest.get("ordered_extensions", [])
        ]
        extension_entries = {
            str(entry.get("id", "")): entry for entry in manifest.get("extensions", [])
        }
        assert extension_id in set(ordered_extensions)

        if pinned:
            assert str(extension_entries.get(extension_id, {}).get("version", "")) == (
                resolved_version
            )

        if not code_binary:
            return

        with _misconfigured_proxy(monkeypatch):
            extension_manager_module.import_offline_bundle(
                bundle_path=str(bundle_path),
                target_path=str(tmp_path / "import-cache"),
                code_path=code_binary,
                strict_offline=True,
            )

        assert extension_id in set(ordered_extensions)
    finally:
        if code_binary:
            for installed_extension_id in set(ordered_extensions + [extension_id]):
                _uninstall_extension(code_binary, installed_extension_id)


@pytest.mark.slow
def test_integration_offline_bundle_import_local_vscode_install_with_misconfigured_proxy(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    if os.environ.get("GITHUB_ACTIONS", "").lower() == "true":
        pytest.skip("Local VS Code CLI install validation is skipped on CI")

    code_binary = shutil.which("code")
    if not code_binary:
        pytest.skip("VS Code CLI not available")

    extension_id = "dbaeumer.vscode-eslint"
    config_path = tmp_path / "single-extension.json"
    config_path.write_text(
        json.dumps(
            {
                "customizations": {
                    "vscode": {
                        "extensions": [extension_id],
                    }
                }
            }
        ),
        encoding="utf-8",
    )

    bundle_path = tmp_path / "bundle-local"
    extension_manager_module.export_offline_bundle(
        config_name=str(config_path),
        bundle_path=str(bundle_path),
        target_path=str(_artifact_cache_dir(tmp_path, "downloads")),
        code_path=code_binary,
    )

    sandbox_vscode_root = tmp_path / "vscode-server-local"
    sandbox_extensions = sandbox_vscode_root / "extensions"
    sandbox_extensions.mkdir(parents=True, exist_ok=True)
    (sandbox_extensions / "extensions.json").write_text("[]", encoding="utf-8")
    monkeypatch.setattr(extension_manager_module, "vscode_root", sandbox_vscode_root)

    isolated_code_binary = _isolated_code_binary(
        code_binary, tmp_path / "isolated-code"
    )

    with _misconfigured_proxy(monkeypatch):
        extension_manager_module.import_offline_bundle(
            bundle_path=str(bundle_path),
            target_path=str(tmp_path / "import-cache-local"),
            code_path=isolated_code_binary,
            strict_offline=True,
        )

    installed = _installed_extensions(isolated_code_binary)
    assert extension_id in installed

    _uninstall_extension(isolated_code_binary, extension_id)
    installed_after_cleanup = _installed_extensions(isolated_code_binary)
    assert extension_id not in installed_after_cleanup



from __future__ import annotations

import os
import platform
from pathlib import Path


def detect_runtime_environment() -> str:
    """Return a simple runtime classification used by path and CLI discovery."""
    explicit_environment = os.environ.get("UVSCEM_RUNTIME", "").strip().lower()
    if explicit_environment:
        return explicit_environment

    if os.environ.get("VSCODE_AGENT_FOLDER", "").strip():
        return "vscode-remote"

    if Path.home().joinpath(".vscode-server").exists():
        return "vscode-server"

    if Path.home().joinpath(".vscode-remote").exists():
        return "vscode-remote"

    return "local"


def resolve_vscode_root() -> Path:
    """Resolve the VS Code data root with Linux/devcontainer-first defaults."""
    explicit_root = os.environ.get("UVSCEM_VSCODE_ROOT", "").strip()
    if explicit_root:
        return Path(explicit_root).expanduser().resolve()

    agent_root = os.environ.get("VSCODE_AGENT_FOLDER", "").strip()
    if agent_root:
        return Path(agent_root).expanduser().resolve()

    runtime_environment = detect_runtime_environment()
    if runtime_environment == "vscode-server":
        return Path.home().joinpath(".vscode-server").resolve()
    if runtime_environment == "vscode-remote":
        return Path.home().joinpath(".vscode-remote").resolve()

    home = Path.home()
    candidates = [
        home.joinpath(".vscode-server"),
        home.joinpath(".vscode-remote"),
        home.joinpath(".vscode"),
    ]
    for candidate in candidates:
        if candidate.exists():
            return candidate.resolve()

    if platform.system().lower() == "linux":
        return home.joinpath(".vscode-server").resolve()
    return home.joinpath(".vscode").resolve()



#! /bin/env python3
from __future__ import annotations

import argparse
import asyncio
import json
import logging
import os
import shutil
import subprocess
import tempfile
import zipfile
from collections import defaultdict
from pathlib import Path
from typing import Any

# for parsing devcontainer.json (if it includes comments etc.)
import json5
import requests
from dependency_algorithm import Dependencies

from uvscem.api_client import CodeAPIManager
from uvscem.bundle_io import (
    build_bundle_manifest,
    build_extension_manifest_entry,
    resolve_bundled_vsce_sign_binary,
    resolve_vsce_sign_targets,
    sha256_file,
)
from uvscem.code_manager import CodeManager
from uvscem.exceptions import (
    InstallationWorkflowError,
    OfflineBundleExportError,
    OfflineBundleImportExecutionError,
    OfflineBundleImportMissingFileError,
    OfflineBundleImportValidationError,
    OfflineModeError,
)
from uvscem.install_engine import (
    run_code_cli_install,
    run_vsce_sign_verify,
    stream_download_to_target,
)
from uvscem.internal_config import (
    DEFAULT_USER_AGENT,
    HTTP_STREAM_CONNECT_TIMEOUT_SECONDS,
    HTTP_STREAM_READ_TIMEOUT_SECONDS,
    MAX_INSTALL_RETRIES,
)
from uvscem.models import ExtensionSpec, ResolvedExtensionRequest
from uvscem.vsce_sign_bootstrap import (
    DEFAULT_VSCE_SIGN_VERSION,
    SUPPORTED_VSCE_SIGN_TARGETS,
    get_vsce_sign_package_name,
    get_vsce_sign_target,
    install_vsce_sign_binary_for_target,
    provision_vsce_sign_binary_for_run,
)
from uvscem.vscode_paths import resolve_vscode_root

__author__ = "Arne Fahrenwalde <arne@fahrenwal.de>"


# attempt to install an extension a maximum of three times
max_retries = MAX_INSTALL_RETRIES
user_agent: str = DEFAULT_USER_AGENT
# VSCode extension installation directory
vscode_root: Path = resolve_vscode_root()
logger: logging.Logger = logging.getLogger(__name__)


def _workflow_error_message(operation: str, exc: Exception) -> str:
    return f"{operation} failed: {exc}"


def _write_json_atomic(path: Path, payload: Any) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    with tempfile.NamedTemporaryFile(
        mode="w",
        encoding="utf-8",
        dir=str(path.parent),
        prefix=f"{path.name}.",
        suffix=".tmp",
        delete=False,
    ) as tmp_file:
        tmp_path = Path(tmp_file.name)
        json.dump(payload, tmp_file)
        tmp_file.flush()
        os.fsync(tmp_file.fileno())
    tmp_path.replace(path)


class CodeExtensionManager(object):
    """Proxy aware helper script to install VSCode extensions in a DevContainer."""

    api_manager: CodeAPIManager
    socket_manager: CodeManager
    code_binary: str
    dev_container_config_path: Path
    extensions: list[str]
    installed: list[dict[str, Any]]
    extension_dependencies: dict[str, set[str]]
    extension_metadata: dict[str, dict[str, Any]]
    target_path: Path
    vsce_sign_binary: Path | None

    def __init__(
        self,
        config_name: str = "devcontainer.json",
        code_path: str = "code",
        target_directory: str = "",
    ) -> None:
        """Set up the CodeExtensionManager."""
        self.extension_dependencies = defaultdict(set)
        self.extension_metadata = {}
        self.api_manager = CodeAPIManager()
        self.socket_manager = CodeManager()
        self.code_binary = code_path
        config_path = Path(config_name)
        self.dev_container_config_path = (
            config_path
            if config_path.is_absolute()
            else Path.cwd()
            .joinpath(
                config_path,
            )
            .absolute()
        )
        self.extensions = []
        self.installed = []
        self.target_path = (
            Path(target_directory).absolute()
            if target_directory
            else Path.home().joinpath("cache/.vscode/extensions").absolute()
        )
        self.vsce_sign_binary = None
        # create target directory if necessary
        self.target_path.mkdir(parents=True, exist_ok=True)

    async def initialize(self) -> None:
        """Asynchronously initialize runtime state before installation."""
        await self.socket_manager.initialize()
        self.extensions = await self.parse_all_extensions()
        self.installed = await self.find_installed()

    def sanitize_dependencies(self, extensions: list[str]) -> list[str]:
        """
        Remove VSCode internal extensions which are not available
        on the Marketplace, but still are listed as dependency.
        """
        sanitized = []

        for e in extensions:
            if not e.startswith("vscode."):
                sanitized.append(e)

        return sanitized

    def add_missing_dependency(
        self, extensions: list[str], dependencies: list[str]
    ) -> None:
        for dependency in dependencies:
            if dependency not in extensions:
                extensions.append(dependency)

    async def _load_extension_specs(self) -> list[ExtensionSpec]:
        parsed_config: dict[str, Any] = await asyncio.to_thread(
            lambda: json5.loads(self.dev_container_config_path.read_text())
        )
        raw_extensions: list[str] = list(
            parsed_config.get("customizations", {})
            .get("vscode", {})
            .get("extensions", [])
        )
        return [
            ExtensionSpec(*self.parse_extension_spec(extension_spec))
            for extension_spec in raw_extensions
        ]

    async def _resolve_extension_requests(
        self, specs: list[ExtensionSpec]
    ) -> tuple[list[str], dict[str, str]]:
        extensions: list[str] = []
        extension_pins: dict[str, str] = {}
        for spec in specs:
            if not spec.extension_id:
                continue
            extensions.append(spec.extension_id)
            if spec.pinned_version:
                extension_pins[spec.extension_id] = spec.pinned_version
        return extensions, extension_pins

    async def _fetch_metadata_for_request(
        self,
        request: ResolvedExtensionRequest,
    ) -> dict[str, Any]:
        metadata = await self.api_manager.get_extension_metadata(
            request.extension_id,
            include_latest_stable_version_only=not bool(request.pinned_version),
            requested_version=request.pinned_version,
        )
        versions = metadata.get(request.extension_id, [])
        if not versions:
            if request.pinned_version:
                raise ValueError(
                    f"Pinned extension version not found: {request.extension_id}@{request.pinned_version}"
                )
            return {}
        return dict(versions[0])

    async def parse_all_extensions(self) -> list[str]:
        specs = await self._load_extension_specs()
        extensions, extension_pins = await self._resolve_extension_requests(specs)

        for extension_id in extensions:
            request = ResolvedExtensionRequest(
                extension_id=extension_id,
                pinned_version=extension_pins.get(extension_id, ""),
            )
            metadata = await self._fetch_metadata_for_request(request)
            if not metadata:
                continue
            self.extension_metadata[extension_id] = metadata

            dependencies = self.sanitize_dependencies(metadata.get("dependencies", []))
            self.extension_dependencies[extension_id].update(dependencies)
            self.add_missing_dependency(extensions, dependencies)

            # dont treat extension packs as dependencies
            extension_pack = self.sanitize_dependencies(
                metadata.get("extension_pack", [])
            )
            self.add_missing_dependency(extensions, extension_pack)

        # get installation order
        dependencies = Dependencies(self.extension_dependencies)
        return dependencies.resolve_dependencies()

    def parse_extension_spec(self, extension_spec: str) -> tuple[str, str]:
        spec = extension_spec.strip()
        if not spec:
            return "", ""
        if "@" not in spec:
            return spec, ""
        extension_id, _, requested_version = spec.rpartition("@")
        extension_id = extension_id.strip()
        requested_version = requested_version.strip()
        if not extension_id or not requested_version:
            return spec, ""
        return extension_id, requested_version

    def get_dirname(self, extension_id: str) -> str:
        version: str = str(
            self.extension_metadata.get(extension_id, {}).get("version", "")
        )
        return f"{extension_id}-{version}"

    def get_filename(self, extension_id: str) -> str:
        return f"{self.get_dirname(extension_id)}.vsix"

    def get_signature_filename(self, extension_id: str) -> str:
        return f"{self.get_dirname(extension_id)}.sigzip"

    async def download_extension(self, extension_id: str) -> Path:
        """Download an extension and return the downloaded file's path."""
        metadata: dict = self.extension_metadata.get(extension_id, {})
        url: str = str(metadata.get("url", ""))
        if not url:
            raise ValueError(f"Missing download URL for extension: {extension_id}")
        headers: dict = {
            "User-Agent": user_agent,
        }

        def _download_sync() -> Path:
            target_path = self.target_path.joinpath(self.get_filename(extension_id))
            if target_path.is_file() and target_path.stat().st_size > 0:
                logger.info(f"Reusing cached VSIX for {extension_id}: {target_path}")
                return target_path
            logger.info(f"Installing {extension_id} from {url}")
            return stream_download_to_target(
                session=self.api_manager.session,
                url=url,
                target_path=target_path,
                headers=headers,
                temp_prefix="vscode-extension.",
                timeout=(
                    HTTP_STREAM_CONNECT_TIMEOUT_SECONDS,
                    HTTP_STREAM_READ_TIMEOUT_SECONDS,
                ),
            )

        return await asyncio.to_thread(_download_sync)

    async def download_signature_archive(self, extension_id: str) -> Path:
        """Download an extension signature archive and return the downloaded file path."""
        metadata: dict = self.extension_metadata.get(extension_id, {})
        url: str = str(metadata.get("signature", ""))
        if not url:
            raise ValueError(f"Missing signature URL for extension: {extension_id}")
        headers: dict = {
            "User-Agent": user_agent,
        }

        def _download_sync() -> Path:
            target_path = self.target_path.joinpath(
                self.get_signature_filename(extension_id)
            )
            if target_path.is_file() and target_path.stat().st_size > 0:
                logger.info(
                    f"Reusing cached signature archive for {extension_id}: {target_path}"
                )
                return target_path
            return stream_download_to_target(
                session=self.api_manager.session,
                url=url,
                target_path=target_path,
                headers=headers,
                temp_prefix="vscode-extension-signature.",
                timeout=(
                    HTTP_STREAM_CONNECT_TIMEOUT_SECONDS,
                    HTTP_STREAM_READ_TIMEOUT_SECONDS,
                ),
            )

        return await asyncio.to_thread(_download_sync)

    async def verify_extension_signature(
        self, extension_path: Path, signature_archive_path: Path
    ) -> None:
        """Verify one VSIX against its signature archive using vsce-sign."""
        if self.vsce_sign_binary is None:
            raise RuntimeError("vsce-sign binary is not initialized for this run")

        await asyncio.to_thread(
            run_vsce_sign_verify,
            vsce_sign_binary=self.vsce_sign_binary,
            extension_path=extension_path,
            signature_archive_path=signature_archive_path,
            run_command=subprocess.run,
        )

    async def install_extension_manually(
        self, extension_id: str, extension_path: Path, update_json=True
    ) -> None:
        # steps:
        # [x] parse extensions.json for installed extensions -> self.installed
        # [x] unpack everything to ~/.vscode-server/extensions/extension.bundleid-version
        # [x] add installation_metadata to self.installed and export it to ~/.vscode-server/extensions/extensions.json
        # [x] repeat the steps for all extensions listed in extension_pack (check their dependencies!)
        target_path = vscode_root.joinpath(
            f"extensions/{self.get_dirname(extension_id)}"
        )

        def _install_manual_sync() -> None:
            if target_path.exists():
                shutil.rmtree(target_path)

            with tempfile.TemporaryDirectory(
                prefix="vscode-extension-extract."
            ) as tmp_dir:
                with zipfile.ZipFile(extension_path, "r") as zip_obj:
                    zip_obj.extractall(tmp_dir)
                    shutil.move(Path(tmp_dir, "extension/"), target_path)

        await asyncio.to_thread(_install_manual_sync)

        if update_json:
            await self.update_extensions_json(extension_id=extension_id)

    async def update_extensions_json(
        self, extension_id: str = "", extension_ids: list[str] | None = None
    ):
        self.installed = await self.find_installed()
        extension_ids = list(extension_ids or [])

        if extension_id:
            extension_ids.append(extension_id)

        def _metadata_extension_id(metadata: dict[str, Any]) -> str:
            return str(metadata.get("identifier", {}).get("id", "")).lower()

        for eid in extension_ids:
            installation_metadata = self.extension_metadata.get(eid, {}).get(
                "installation_metadata"
            )
            if installation_metadata:
                incoming = dict(installation_metadata)
                incoming_id = _metadata_extension_id(incoming)
                if not incoming_id:
                    self.installed.append(incoming)
                    continue

                replaced = False
                deduplicated: list[dict[str, Any]] = []
                for item in self.installed:
                    if _metadata_extension_id(item) == incoming_id:
                        if not replaced:
                            deduplicated.append(incoming)
                            replaced = True
                        continue
                    deduplicated.append(item)

                if not replaced:
                    deduplicated.append(incoming)

                self.installed = deduplicated

        def _update_sync() -> None:
            json_path: Path = self.extensions_json_path()
            _write_json_atomic(json_path, self.installed)

        await asyncio.to_thread(_update_sync)

    async def install_extension(
        self, extension_id: str, extension_pack: bool = False, retries: int = 0
    ) -> None:
        """Install a single extension."""
        file_name = self.get_filename(extension_id)
        file_path = self.target_path.joinpath(file_name)
        metadata = self.extension_metadata.get(extension_id, {})

        if file_path.is_file():
            logger.info(f"File {file_name} exists - skipping download...")
        else:
            file_path = await self.download_extension(extension_id)

        signature_url: str = str(metadata.get("signature", ""))
        if signature_url:
            signature_file = self.target_path.joinpath(
                self.get_signature_filename(extension_id)
            )
            if not signature_file.is_file():
                signature_file = await self.download_signature_archive(extension_id)
            await self.verify_extension_signature(file_path, signature_file)
        else:
            logger.warning(
                f"Missing signature metadata for extension {extension_id}, skipping verification"
            )

        # installing extension packs doesn't work because the code install routine tries fetching other packages itself
        extension_pack_items = list(metadata.get("extension_pack", []))
        if extension_pack_items:
            manually_installed = []

            if not extension_pack:
                manually_installed.append(extension_id)
                extension_pack = True

            # also repeat this step for the packages listed in extension pack
            for ep in extension_pack_items:
                if ep in self.extensions:
                    await self.install_extension(
                        self.extensions.pop(self.extensions.index(ep)),
                        extension_pack=True,
                    )
                manually_installed.append(ep)

            await self.update_extensions_json(extension_ids=manually_installed)

        if extension_pack:
            await self.install_extension_manually(
                extension_id, file_path, update_json=False
            )
        else:
            try:
                await asyncio.to_thread(
                    run_code_cli_install,
                    code_binary=self.code_binary,
                    extension_path=file_path,
                    run_command=subprocess.run,
                )
            except subprocess.CalledProcessError as e:
                logger.error(f"Something went wrong: {e}")
                await self.socket_manager.find_socket(update_environment=True)
                if retries < max_retries:
                    await self.install_extension(
                        extension_id,
                        extension_pack=extension_pack,
                        retries=retries + 1,
                    )

    async def find_installed(self) -> list[dict[str, Any]]:
        """Return a list of already installed extensions."""

        def _find_sync() -> list[dict[str, Any]]:
            extensions_path: Path = self.extensions_json_path()
            with open(extensions_path, "r") as f:
                return json.load(f)

        return await asyncio.to_thread(_find_sync)

    def extensions_json_path(self) -> Path:
        return vscode_root.joinpath("extensions/extensions.json")

    async def exclude_installed(self) -> None:
        """Remove all already installed extensions from the extensions list."""
        for installed_extension in self.installed:
            for extension in self.extensions:
                installed_version: str = str(installed_extension.get("version", ""))
                installed_name: str = str(
                    installed_extension.get("identifier", {}).get("id", "")
                ).lower()
                extension_name: str = extension.lower()
                extension_version: str = str(
                    self.extension_metadata.get(extension, {}).get("version", "")
                )
                if (
                    extension_name == installed_name
                    and installed_version == extension_version
                ):
                    self.extensions.remove(extension)
                    logger.info(
                        f"Skipping {extension} ({extension_version}), already installed."
                    )

    async def install_async(self) -> None:
        """Asynchronously install all configured extensions."""
        await self.exclude_installed()

        provisioner = provision_vsce_sign_binary_for_run(install_dir=self.target_path)
        with provisioner as vsce_sign_binary:
            self.vsce_sign_binary = vsce_sign_binary
            try:
                while self.extensions:
                    await self.install_extension(self.extensions.pop(0))
                    await asyncio.sleep(1)
            finally:
                self.vsce_sign_binary = None


def install(
    config_name: str = "devcontainer.json",
    code_path: str = "code",
    target_path: str = "$HOME/cache/.vscode/extensions",
    log_level: str = "info",
) -> None:
    """Install all extensions listed in devcontainer.json."""
    logging.basicConfig(
        level=(getattr(logging, log_level.upper())),
        format="%(relativeCreated)d [%(levelname)s] %(message)s",
    )
    logger.info("Attempting to install all necessary DevContainer extensions.")

    async def _run() -> None:
        manager = CodeExtensionManager(
            config_name=config_name,
            code_path=code_path,
            target_directory=target_path,
        )
        await manager.initialize()
        await manager.install_async()

    try:
        asyncio.run(_run())
    except (
        requests.RequestException,
        subprocess.CalledProcessError,
        OSError,
        RuntimeError,
        ValueError,
    ) as exc:
        raise InstallationWorkflowError(
            _workflow_error_message("Extension installation", exc)
        ) from exc


def _run_command(cmd: list[str]) -> None:
    process = subprocess.run(
        cmd,
        capture_output=True,
        check=False,
        text=True,
    )
    if process.returncode != 0:
        raise subprocess.CalledProcessError(
            process.returncode,
            cmd,
            process.stdout,
            process.stderr,
        )


def _sign_bundle_manifest(manifest_path: Path, signing_key: str) -> Path:
    signature_path = manifest_path.with_suffix(f"{manifest_path.suffix}.asc")
    _run_command(
        [
            "gpg",
            "--batch",
            "--yes",
            "--armor",
            "--local-user",
            signing_key,
            "--output",
            str(signature_path),
            "--detach-sign",
            str(manifest_path),
        ]
    )
    return signature_path


def _verify_bundle_manifest_signature(
    manifest_path: Path,
    signature_path: Path,
    verification_keyring: str = "",
) -> None:
    cmd = ["gpg", "--batch"]
    if verification_keyring:
        cmd.extend(
            [
                "--no-default-keyring",
                "--keyring",
                verification_keyring,
            ]
        )
    cmd.extend(["--verify", str(signature_path), str(manifest_path)])
    _run_command(cmd)


def export_offline_bundle(
    config_name: str = "devcontainer.json",
    bundle_path: str = "./uvscem-offline-bundle",
    target_path: str = "$HOME/cache/.vscode/extensions",
    code_path: str = "code",
    log_level: str = "info",
    vsce_sign_version: str = DEFAULT_VSCE_SIGN_VERSION,
    vsce_sign_targets: str = "current",
    manifest_signing_key: str = "",
) -> None:
    """Export extensions, signatures, and vsce-sign into an offline bundle."""
    logging.basicConfig(
        level=(getattr(logging, log_level.upper())),
        format="%(relativeCreated)d [%(levelname)s] %(message)s",
    )

    try:
        bundle_dir = Path(bundle_path).expanduser().resolve()
        artifacts_dir = bundle_dir.joinpath("artifacts")
        cache_dir = Path(os.path.expandvars(target_path)).expanduser().resolve()
        vsce_sign_dir = bundle_dir.joinpath("vsce-sign")
        bundle_dir.mkdir(parents=True, exist_ok=True)
        artifacts_dir.mkdir(parents=True, exist_ok=True)
        cache_dir.mkdir(parents=True, exist_ok=True)
        vsce_sign_dir.mkdir(parents=True, exist_ok=True)

        resolved_targets = resolve_vsce_sign_targets(
            value=vsce_sign_targets,
            current_target=get_vsce_sign_target(),
            supported_targets=SUPPORTED_VSCE_SIGN_TARGETS,
        )

        async def _run() -> tuple[
            list[str],
            dict[str, dict[str, Any]],
            dict[str, tuple[Path, Path]],
        ]:
            manager = CodeExtensionManager(
                config_name=config_name,
                code_path=code_path,
                target_directory=str(cache_dir),
            )
            extensions = await manager.parse_all_extensions()
            downloaded: dict[str, tuple[Path, Path]] = {}
            for extension_id in extensions:
                vsix_path = await manager.download_extension(extension_id)
                signature_path = await manager.download_signature_archive(extension_id)
                downloaded[extension_id] = (vsix_path, signature_path)
            return extensions, manager.extension_metadata, downloaded

        extensions, extension_metadata, downloaded_artifacts = asyncio.run(_run())

        for extension_id in extensions:
            vsix_path, signature_path = downloaded_artifacts[extension_id]
            shutil.copy2(vsix_path, artifacts_dir.joinpath(vsix_path.name))
            shutil.copy2(signature_path, artifacts_dir.joinpath(signature_path.name))

        vsce_sign_binaries: list[dict[str, str]] = []
        for target in resolved_targets:
            target_dir = vsce_sign_dir.joinpath(target)
            binary_path = install_vsce_sign_binary_for_target(
                target=target,
                install_dir=target_dir,
                version=vsce_sign_version,
                force=True,
            )
            vsce_sign_binaries.append(
                {
                    "target": target,
                    "package": get_vsce_sign_package_name(target),
                    "binary": str(binary_path.relative_to(bundle_dir)),
                    "sha256": sha256_file(binary_path),
                }
            )

        extension_entries = [
            build_extension_manifest_entry(
                extension_id=extension_id,
                metadata=extension_metadata.get(extension_id, {}),
                artifacts_dir=artifacts_dir,
            )
            for extension_id in extensions
        ]

        manifest = build_bundle_manifest(
            config_name=config_name,
            ordered_extensions=extensions,
            extension_entries=extension_entries,
            vsce_sign_version=vsce_sign_version,
            vsce_sign_binaries=vsce_sign_binaries,
        )
        manifest_path = bundle_dir.joinpath("manifest.json")
        manifest_path.write_text(
            json.dumps(manifest, indent=2, sort_keys=True),
            encoding="utf-8",
        )
        if manifest_signing_key:
            _sign_bundle_manifest(manifest_path, manifest_signing_key)
    except (
        requests.RequestException,
        subprocess.CalledProcessError,
        OSError,
        RuntimeError,
        ValueError,
    ) as exc:
        raise OfflineBundleExportError(
            _workflow_error_message("Offline bundle export", exc)
        ) from exc


def import_offline_bundle(
    bundle_path: str,
    code_path: str = "code",
    target_path: str = "$HOME/cache/.vscode/extensions",
    log_level: str = "info",
    strict_offline: bool = False,
    verify_manifest_signature: bool = False,
    manifest_verification_keyring: str = "",
) -> None:
    """Install extensions from a previously exported offline bundle."""
    logging.basicConfig(
        level=(getattr(logging, log_level.upper())),
        format="%(relativeCreated)d [%(levelname)s] %(message)s",
    )

    try:
        bundle_dir = Path(bundle_path).expanduser().resolve()
        manifest_path = bundle_dir.joinpath("manifest.json")
        if not manifest_path.is_file():
            raise FileNotFoundError(
                f"Offline bundle manifest not found: {manifest_path}"
            )

        if verify_manifest_signature:
            signature_path = manifest_path.with_suffix(f"{manifest_path.suffix}.asc")
            if not signature_path.is_file():
                raise FileNotFoundError(
                    f"Offline bundle manifest signature not found: {signature_path}"
                )
            _verify_bundle_manifest_signature(
                manifest_path=manifest_path,
                signature_path=signature_path,
                verification_keyring=manifest_verification_keyring,
            )

        manifest = json.loads(manifest_path.read_text(encoding="utf-8"))
        if int(manifest.get("schema_version", 0)) != 1:
            raise ValueError("Unsupported offline bundle schema version")

        manager = CodeExtensionManager(
            config_name=str(manifest_path),
            code_path=code_path,
            target_directory=target_path,
        )

        if strict_offline and hasattr(manager, "api_manager"):
            session = getattr(manager.api_manager, "session", None)
            if session is not None:

                def _offline_request(*_args, **_kwargs):
                    raise RuntimeError(
                        "Network access is disabled in strict offline mode"
                    )

                setattr(session, "request", _offline_request)

        extensions_json = manager.extensions_json_path()
        extensions_json.parent.mkdir(parents=True, exist_ok=True)
        if not extensions_json.is_file():
            extensions_json.write_text("[]", encoding="utf-8")

        extension_entries = {
            str(entry.get("id", "")): entry for entry in manifest.get("extensions", [])
        }
        ordered_extensions: list[str] = [
            str(extension_id) for extension_id in manifest.get("ordered_extensions", [])
        ]

        artifacts_dir = bundle_dir.joinpath("artifacts")
        for extension_id in ordered_extensions:
            entry = extension_entries.get(extension_id)
            if not entry:
                raise ValueError(
                    f"Missing extension manifest entry for extension: {extension_id}"
                )

            filename = str(entry.get("filename", ""))
            signature_filename = str(entry.get("signature_filename", ""))
            source_vsix = artifacts_dir.joinpath(filename)
            source_signature = artifacts_dir.joinpath(signature_filename)
            if not source_vsix.is_file() or not source_signature.is_file():
                raise FileNotFoundError(
                    f"Missing offline artifact(s) for {extension_id}: {filename}, {signature_filename}"
                )

            expected_vsix_sha256 = str(entry.get("vsix_sha256", ""))
            if (
                expected_vsix_sha256
                and sha256_file(source_vsix) != expected_vsix_sha256
            ):
                raise ValueError(f"VSIX checksum mismatch for {extension_id}")

            expected_signature_sha256 = str(entry.get("signature_sha256", ""))
            if (
                expected_signature_sha256
                and sha256_file(source_signature) != expected_signature_sha256
            ):
                raise ValueError(f"Signature checksum mismatch for {extension_id}")

            shutil.copy2(source_vsix, manager.target_path.joinpath(filename))
            shutil.copy2(
                source_signature,
                manager.target_path.joinpath(signature_filename),
            )

            manager.extension_metadata[extension_id] = {
                "version": str(entry.get("version", "")),
                "url": "offline",
                "signature": "offline",
                "installation_metadata": entry.get("installation_metadata", {}),
                "dependencies": list(entry.get("dependencies", [])),
                "extension_pack": list(entry.get("extension_pack", [])),
            }

        vsce_sign_info = manifest.get("vsce_sign", {})
        vsce_sign_binary, expected_vsce_sign_sha256 = resolve_bundled_vsce_sign_binary(
            bundle_dir=bundle_dir,
            vsce_sign_info=vsce_sign_info,
            current_target=get_vsce_sign_target(),
        )

        if not vsce_sign_binary.is_file():
            raise FileNotFoundError(
                f"Bundled vsce-sign binary not found: {vsce_sign_binary}"
            )
        if (
            expected_vsce_sign_sha256
            and sha256_file(vsce_sign_binary) != expected_vsce_sign_sha256
        ):
            raise ValueError("Bundled vsce-sign checksum mismatch")

        async def _run() -> None:
            manager.installed = await manager.find_installed()
            manager.extensions = list(ordered_extensions)
            await manager.exclude_installed()

            manager.vsce_sign_binary = vsce_sign_binary
            try:
                while manager.extensions:
                    await manager.install_extension(manager.extensions.pop(0))
                    await asyncio.sleep(1)
            finally:
                manager.vsce_sign_binary = None

        asyncio.run(_run())
    except FileNotFoundError as exc:
        raise OfflineBundleImportMissingFileError(
            _workflow_error_message("Offline bundle import", exc)
        ) from exc
    except ValueError as exc:
        raise OfflineBundleImportValidationError(
            _workflow_error_message("Offline bundle import", exc)
        ) from exc
    except RuntimeError as exc:
        if "strict offline mode" in str(exc):
            raise OfflineModeError(
                _workflow_error_message("Offline bundle import", exc)
            ) from exc
        raise OfflineBundleImportExecutionError(
            _workflow_error_message("Offline bundle import", exc)
        ) from exc
    except (requests.RequestException, subprocess.CalledProcessError, OSError) as exc:
        raise OfflineBundleImportExecutionError(
            _workflow_error_message("Offline bundle import", exc)
        ) from exc


def build_parser() -> argparse.ArgumentParser:
    """Create CLI parser while keeping existing install option names."""
    parser = argparse.ArgumentParser(prog="uvscem")
    parser.add_argument(
        "command",
        nargs="?",
        default="install",
        choices=["install", "export", "import"],
    )
    parser.add_argument("--config-name", default="devcontainer.json")
    parser.add_argument("--code-path", default="code")
    parser.add_argument("--target-path", default="$HOME/cache/.vscode/extensions")
    parser.add_argument("--bundle-path", default="./uvscem-offline-bundle")
    parser.add_argument("--vsce-sign-version", default=DEFAULT_VSCE_SIGN_VERSION)
    parser.add_argument("--vsce-sign-targets", default="current")
    parser.add_argument("--manifest-signing-key", default="")
    parser.add_argument("--strict-offline", action="store_true")
    parser.add_argument("--verify-manifest-signature", action="store_true")
    parser.add_argument("--manifest-verification-keyring", default="")
    parser.add_argument("--log-level", default="info")
    return parser


def main() -> None:
    parser = build_parser()
    args = parser.parse_args()
    command = getattr(args, "command", "install")

    if command == "install":
        install(
            config_name=args.config_name,
            code_path=args.code_path,
            target_path=args.target_path,
            log_level=args.log_level,
        )
        return

    if command == "export":
        export_offline_bundle(
            config_name=args.config_name,
            bundle_path=args.bundle_path,
            target_path=args.target_path,
            code_path=args.code_path,
            log_level=args.log_level,
            vsce_sign_version=args.vsce_sign_version,
            vsce_sign_targets=getattr(args, "vsce_sign_targets", "current"),
            manifest_signing_key=getattr(args, "manifest_signing_key", ""),
        )
        return

    import_offline_bundle(
        bundle_path=args.bundle_path,
        code_path=args.code_path,
        target_path=args.target_path,
        log_level=args.log_level,
        strict_offline=getattr(args, "strict_offline", False),
        verify_manifest_signature=getattr(args, "verify_manifest_signature", False),
        manifest_verification_keyring=getattr(
            args, "manifest_verification_keyring", ""
        ),
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

import importlib
import sys
import tempfile
import types
from pathlib import Path

import pytest


class _ManagerStub:
    def __init__(
        self, config_name: str, code_path: str, target_directory: str = ""
    ) -> None:
        self.config_name = config_name
        self.code_path = code_path
        self.target_directory = target_directory
        self.install_called = False
        self.initialized = False

    def install(self) -> None:
        self.install_called = True

    async def install_async(self) -> None:
        self.install_called = True

    async def initialize(self) -> None:
        self.initialized = True


def _import_extension_manager(monkeypatch: pytest.MonkeyPatch):
    if "json5" not in sys.modules:
        json5_stub = types.ModuleType("json5")
        setattr(json5_stub, "loads", lambda _value: {})
        monkeypatch.setitem(sys.modules, "json5", json5_stub)

    if "requests" not in sys.modules:
        requests_stub = types.ModuleType("requests")
        setattr(requests_stub, "Response", object)
        monkeypatch.setitem(sys.modules, "requests", requests_stub)

    if "dependency_algorithm" not in sys.modules:
        dependency_stub = types.ModuleType("dependency_algorithm")

        class _Dependencies:
            def __init__(self, _deps) -> None:
                self._deps = _deps

            def resolve_dependencies(self):
                return []

        setattr(dependency_stub, "Dependencies", _Dependencies)
        monkeypatch.setitem(sys.modules, "dependency_algorithm", dependency_stub)

    if "uvscem.api_client" not in sys.modules:
        api_stub = types.ModuleType("uvscem.api_client")

        class _CodeAPIManager:
            pass

        setattr(api_stub, "CodeAPIManager", _CodeAPIManager)
        monkeypatch.setitem(sys.modules, "uvscem.api_client", api_stub)

    if "uvscem.code_manager" not in sys.modules:
        code_stub = types.ModuleType("uvscem.code_manager")

        class _CodeManager:
            pass

        setattr(code_stub, "CodeManager", _CodeManager)
        monkeypatch.setitem(sys.modules, "uvscem.code_manager", code_stub)

    if "uvscem.extension_manager" in sys.modules:
        monkeypatch.delitem(sys.modules, "uvscem.extension_manager", raising=False)

    return importlib.import_module("uvscem.extension_manager")


def _run_main(
    monkeypatch: pytest.MonkeyPatch,
    args: list[str],
    extension_manager=None,
) -> int:
    if extension_manager is None:
        extension_manager = _import_extension_manager(monkeypatch)
    monkeypatch.setattr(sys, "argv", ["uvscem", *args])
    try:
        extension_manager.main()
    except SystemExit as exc:  # pragma: no cover - framework dependent
        code = exc.code
        if isinstance(code, int):
            return code
        return 1
    return 0


def test_cli_help_shows_core_options(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
) -> None:
    code = _run_main(monkeypatch, ["--help"])
    output = capsys.readouterr()
    help_text = f"{output.out}\n{output.err}"

    assert code == 0
    assert "--config-name" in help_text
    assert "--code-path" in help_text
    assert "--target-path" in help_text
    assert "--log-level" in help_text


def test_cli_invokes_manager_with_expected_arguments(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    extension_manager = _import_extension_manager(monkeypatch)
    captured: dict[str, object] = {}
    code_path = str(Path(tempfile.gettempdir()) / "custom" / "code")
    target_path = str(Path(tempfile.gettempdir()) / "extensions")

    def _factory(
        config_name: str, code_path: str, target_directory: str = ""
    ) -> _ManagerStub:
        manager = _ManagerStub(
            config_name=config_name,
            code_path=code_path,
            target_directory=target_directory,
        )
        captured["manager"] = manager
        return manager

    monkeypatch.setattr(extension_manager, "CodeExtensionManager", _factory)

    code = _run_main(
        monkeypatch,
        [
            "--config-name",
            "custom-devcontainer.json",
            "--code-path",
            code_path,
            "--target-path",
            target_path,
            "--log-level",
            "info",
        ],
        extension_manager=extension_manager,
    )

    assert code == 0
    manager = captured["manager"]
    assert isinstance(manager, _ManagerStub)
    assert manager.config_name == "custom-devcontainer.json"
    assert manager.code_path == code_path
    assert manager.target_directory == target_path
    assert manager.initialized is True
    assert manager.install_called is True


def test_cli_reports_argument_errors(
    monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
) -> None:
    code = _run_main(monkeypatch, ["--not-a-real-option"])
    output = capsys.readouterr()
    error_text = f"{output.out}\n{output.err}".lower()

    assert code != 0
    assert "error" in error_text



from __future__ import annotations

import pytest


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption(
        "--slow",
        action="store_true",
        default=False,
        help="run only tests marked as slow",
    )


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line("markers", "slow: marks tests as slow")


def pytest_collection_modifyitems(
    config: pytest.Config, items: list[pytest.Item]
) -> None:
    run_slow = bool(config.getoption("--slow"))

    if run_slow:
        selected: list[pytest.Item] = []
        deselected: list[pytest.Item] = []
        for item in items:
            if "slow" in item.keywords:
                selected.append(item)
            else:
                deselected.append(item)

        if deselected:
            config.hook.pytest_deselected(items=deselected)
        items[:] = selected
        return

    skip_slow = pytest.mark.skip(reason="need --slow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)



from __future__ import annotations

import datetime
from pathlib import Path
from typing import Any

FLAG_INCLUDE_VERSIONS = 0x1
FLAG_INCLUDE_FILES = 0x2
FLAG_INCLUDE_CATEGORY_AND_TAGS = 0x4
FLAG_INCLUDE_SHARED_ACCOUNTS = 0x8
FLAG_INCLUDE_VERSION_PROPERTIES = 0x10
FLAG_EXCLUDE_NON_VALIDATED = 0x20
FLAG_INCLUDE_INSTALLATION_TARGETS = 0x40
FLAG_INCLUDE_ASSET_URI = 0x80
FLAG_INCLUDE_STATISTICS = 0x100
FLAG_INCLUDE_LATEST_VERSION_ONLY = 0x200
FLAG_UNPUBLISHED = 0x1000
FLAG_INCLUDE_NAME_CONFLICT_INFO = 0x8000


def build_query_flags(
    include_versions: bool,
    include_files: bool,
    include_category_and_tags: bool,
    include_shared_accounts: bool,
    include_version_properties: bool,
    exclude_non_validated: bool,
    include_installation_targets: bool,
    include_asset_uri: bool,
    include_statistics: bool,
    include_latest_version_only: bool,
    unpublished: bool,
    include_name_conflict_info: bool,
) -> int:
    flag_map = [
        (include_versions, FLAG_INCLUDE_VERSIONS),
        (include_files, FLAG_INCLUDE_FILES),
        (include_category_and_tags, FLAG_INCLUDE_CATEGORY_AND_TAGS),
        (include_shared_accounts, FLAG_INCLUDE_SHARED_ACCOUNTS),
        (include_version_properties, FLAG_INCLUDE_VERSION_PROPERTIES),
        (exclude_non_validated, FLAG_EXCLUDE_NON_VALIDATED),
        (include_installation_targets, FLAG_INCLUDE_INSTALLATION_TARGETS),
        (include_asset_uri, FLAG_INCLUDE_ASSET_URI),
        (include_statistics, FLAG_INCLUDE_STATISTICS),
        (include_latest_version_only, FLAG_INCLUDE_LATEST_VERSION_ONLY),
        (unpublished, FLAG_UNPUBLISHED),
        (include_name_conflict_info, FLAG_INCLUDE_NAME_CONFLICT_INFO),
    ]
    return sum(flag for enabled, flag in flag_map if enabled)


def build_extension_query_body(
    extension_id: str,
    page_number: int,
    page_size: int,
    flags: int,
) -> dict[str, Any]:
    return {
        "filters": [
            {
                "criteria": [
                    {"filterType": 8, "value": "Microsoft.VisualStudio.Code"},
                    {"filterType": 7, "value": extension_id},
                ],
                "pageNumber": page_number,
                "pageSize": page_size,
                "sortBy": 0,
                "sortOrder": 2,
            }
        ],
        "assetTypes": [],
        "flags": flags,
    }


def shape_extension_metadata_versions(
    extension: dict[str, Any],
    extension_id: str,
    include_latest_stable_version_only: bool,
    requested_version: str,
    vscode_root: Path,
) -> list[dict[str, Any]]:
    result: list[dict[str, Any]] = []
    extension_name: str = str(extension.get("extensionName", ""))
    extensions_versions: list[dict[str, Any]] = list(extension.get("versions", []))
    extensions_statistics: dict[str, Any] = dict(
        {
            (item.get("statisticName"), item.get("value"))
            for item in extension.get("statistics", [])
        }
    )
    extension_publisher_username: str = str(
        extension.get("publisher", {}).get("publisherName", "")
    )

    for extension_version_info in extensions_versions:
        extension_version: str = str(extension_version_info.get("version", ""))
        if requested_version and extension_version != requested_version:
            continue
        extension_properties: dict[str, Any] = dict(
            {
                (item.get("key"), item.get("value"))
                for item in extension_version_info.get("properties", {})
            }
        )
        extension_dependencies: list[str] = [
            item
            for item in str(
                extension_properties.get(
                    "Microsoft.VisualStudio.Code.ExtensionDependencies", ""
                )
            ).split(",")
            if item
        ]
        extension_packs: list[str] = [
            item
            for item in str(
                extension_properties.get(
                    "Microsoft.VisualStudio.Code.ExtensionPack", ""
                )
            ).split(",")
            if item
        ]
        extension_files: dict[str, Any] = dict(
            {
                (item.get("assetType"), item.get("source"))
                for item in extension_version_info.get("files", [])
            }
        )
        extension_download_url: str = str(
            extension_files.get("Microsoft.VisualStudio.Services.VSIXPackage", "")
        )
        extension_signature: str = str(
            extension_files.get("Microsoft.VisualStudio.Services.VsixSignature", "")
        )
        is_pre_release: bool = bool(
            extension_properties.get("Microsoft.VisualStudio.Code.PreRelease", False)
        )

        if is_pre_release and include_latest_stable_version_only:
            continue

        extension_path = vscode_root.joinpath(
            f"extensions/{extension_id}-{extension_version}"
        )
        result.append(
            {
                "publisher": extension_publisher_username,
                "name": extension_name,
                "version": extension_version,
                "dependencies": extension_dependencies,
                "extension_pack": extension_packs,
                "url": extension_download_url,
                "signature": extension_signature,
                "statistics": extensions_statistics,
                "installation_metadata": {
                    "identifier": {"id": extension_id},
                    "version": extension_version,
                    "location": {
                        "$mid": 1,
                        "path": f"{extension_path}",
                        "scheme": "file",
                    },
                    "relativeLocation": f"{extension_id}-{extension_version}",
                    "metadata": {
                        "installedTimestamp": int(
                            (
                                datetime.datetime.now(datetime.timezone.utc)
                                - datetime.datetime(
                                    1970,
                                    1,
                                    1,
                                    tzinfo=datetime.timezone.utc,
                                )
                            ).total_seconds()
                            * 1000
                        ),
                        "id": extension.get("extensionId", ""),
                        "publisherDisplayName": extension.get("publisher", {}).get(
                            "displayName", ""
                        ),
                        "publisherId": extension.get("publisher", {}).get(
                            "publisherId", ""
                        ),
                        "isPreReleaseVersion": is_pre_release,
                    },
                },
            }
        )

        if include_latest_stable_version_only:
            break

    return result
